# encoding=utf8

"""Implementations of Levy function."""

import numpy as np
from niapy.problems.problem import Problem

__all__ = ['Levy']


class Levy(Problem):
    r"""Implementations of Levy functions.

    Date: 2018

    Author: Klemen Berkovič

    License: MIT

    Function:
    **Levy Function**

        :math:`f(\textbf{x}) = \sin^2 (\pi w_1) + \sum_{i = 1}^{D - 1} (w_i - 1)^2 \left( 1 + 10 \sin^2 (\pi w_i + 1) \right) + (w_d - 1)^2 (1 + \sin^2 (2 \pi w_d)) \\ w_i = 1 + \frac{x_i - 1}{4}`

        **Input domain:**
        The function can be defined on any input domain but it is usually
        evaluated on the hypercube :math:`x_i ∈ [-10, 10]`, for all :math:`i = 1, 2,..., D`.

        **Global minimum:**
        :math:`f(\textbf{x}^*) = 0` at :math:`\textbf{x}^* = (1, \cdots, 1)`

    LaTeX formats:
        Inline:
            $f(\textbf{x}) = \sin^2 (\pi w_1) + \sum_{i = 1}^{D - 1} (w_i - 1)^2 \left( 1 + 10 \sin^2 (\pi w_i + 1) \right) + (w_d - 1)^2 (1 + \sin^2 (2 \pi w_d)) \\ w_i = 1 + \frac{x_i - 1}{4}$

        Equation:
            \begin{equation} f(\textbf{x}) = \sin^2 (\pi w_1) + \sum_{i = 1}^{D - 1} (w_i - 1)^2 \left( 1 + 10 \sin^2 (\pi w_i + 1) \right) + (w_d - 1)^2 (1 + \sin^2 (2 \pi w_d)) \\ w_i = 1 + \frac{x_i - 1}{4} \end{equation}

        Domain:
            $-10 \leq x_i \leq 10$

    Reference:
        https://www.sfu.ca/~ssurjano/levy.html

    """

    def __init__(self, dimension=4, lower=-10.0, upper=10.0, *args, **kwargs):
        r"""Initialize Levy problem..

        Args:
            dimension (Optional[int]): Dimension of the problem.
            lower (Optional[Union[float, Iterable[float]]]): Lower bounds of the problem.
            upper (Optional[Union[float, Iterable[float]]]): Upper bounds of the problem.

        See Also:
            :func:`niapy.problems.Problem.__init__`

        """
        super().__init__(dimension, lower, upper, *args, **kwargs)

    @staticmethod
    def latex_code():
        r"""Return the latex code of the problem.

        Returns:
            str: Latex code.

        """
        return r'''$f(\textbf{x}) = \sin^2 (\pi w_1) + \sum_{i = 1}^{D - 1} (w_i - 1)^2 \left( 1 + 10 \sin^2 (\pi w_i + 1) \right) + (w_d - 1)^2 (1 + \sin^2 (2 \pi w_d)) \\ w_i = 1 + \frac{x_i - 1}{4}$'''

    def _evaluate(self, x):
        w = 1 + (x - 1) / 4
        term1 = np.sin(np.pi * w[0]) ** 2
        wi = w[:-1]
        term2 = np.sum((wi - 1) ** 2 * (1 + 10 * np.sin(np.pi * wi + 1)))
        term3 = (w[-1] - 1) ** 2 * (1 + np.sin(2 * np.pi * w[-1]) ** 2)
        return term1 + term2 + term3



# encoding=utf8

"""Implementation of problems utility function."""

from abc import ABC, abstractmethod
import logging
from niapy.util.array import full_array

logging.basicConfig()
logger = logging.getLogger('niapy.problems.problem')
logger.setLevel('INFO')

__all__ = ['Problem']


class Problem(ABC):
    r"""Class representing an optimization problem.

    Attributes:
        dimension (int): Dimension of the problem.
        lower (numpy.ndarray): Lower bounds of the problem.
        upper (numpy.ndarray): Upper bounds of the problem.

    """

    def __init__(self, dimension=1, lower=None, upper=None, *args, **kwargs):
        r"""Initialize Problem.

        Args:
            dimension (Optional[int]): Dimension of the problem.
            lower (Optional[Union[float, Iterable[float]]]): Lower bounds of the problem.
            upper (Optional[Union[float, Iterable[float]]]): Upper bounds of the problem.

        """
        self.dimension = dimension
        self.lower = full_array(lower, dimension)
        self.upper = full_array(upper, dimension)

    @abstractmethod
    def _evaluate(self, x):
        """Evaluate solution."""
        pass

    def evaluate(self, x):
        """Evaluate solution.

        Args:
            x (numpy.ndarray): Solution.

        Returns:
            float: Function value of `x`.

        """
        if x.shape[0] != self.dimension:
            raise ValueError('Dimensions do not match. {} != {}'.format(x.shape[0], self.dimension))

        return self._evaluate(x)

    def __call__(self, x):
        r"""Evaluate solution.

        Args:
            x (numpy.ndarray): Solution.

        Returns:
            float: Function value of `x`.

        See Also:
            :func:`niapy.problems.Problem.evaluate`

        """
        return self.evaluate(x)

    def name(self):
        """Get class name."""
        return self.__class__.__name__



"""Implementation of other algorithms."""

from niapy.algorithms.other.aso import AnarchicSocietyOptimization
from niapy.algorithms.other.hc import HillClimbAlgorithm
from niapy.algorithms.other.mts import MultipleTrajectorySearch, MultipleTrajectorySearchV1, mts_ls1, mts_ls2, mts_ls3, \
    mts_ls1v1, mts_ls3v1
from niapy.algorithms.other.nmm import NelderMeadMethod
from niapy.algorithms.other.rs import RandomSearch
from niapy.algorithms.other.sa import SimulatedAnnealing

__all__ = [
    'NelderMeadMethod',
    'HillClimbAlgorithm',
    'SimulatedAnnealing',
    'MultipleTrajectorySearch',
    'MultipleTrajectorySearchV1',
    'mts_ls1',
    'mts_ls2',
    'mts_ls3',
    'mts_ls1v1',
    'mts_ls3v1',
    'AnarchicSocietyOptimization',
    'RandomSearch'
]



# encoding=utf8
import logging
from math import ceil

import numpy as np

from niapy.algorithms.algorithm import Algorithm, Individual, default_individual_init, default_numpy_init

logging.basicConfig()
logger = logging.getLogger('niapy.algorithms.basic')
logger.setLevel('INFO')

__all__ = ['MonkeyKingEvolutionV1', 'MonkeyKingEvolutionV2', 'MonkeyKingEvolutionV3']


class MkeSolution(Individual):
    r"""Implementation of Monkey King Evolution individual.

    Data:
        2018

    Authors:
        Klemen Berkovič

    License:
        MIT

    Attributes:
        x_pb (array of (float or int)): Personal best position of Monkey particle.
        f_pb (float): Personal best fitness/function value.
        MonkeyKing (bool): Boolean value indicating if particle is Monkey King particle.

    See Also:
        * :class:`niapy.algorithms.Individual`

    """

    def __init__(self, **kwargs):
        r"""Initialize Monkey particle.

        Args:
            **kwargs: Additional arguments

        See Also:
            * :class:`niapy.algorithms.Individual.__init__`

        """
        super().__init__(**kwargs)
        self.f_pb, self.x_pb = self.f, self.x
        self.monkey_king = False

    def update_personal_best(self):
        r"""Update personal best position of particle."""
        if self.f < self.f_pb:
            self.x_pb, self.f_pb = self.x, self.f


class MonkeyKingEvolutionV1(Algorithm):
    r"""Implementation of monkey king evolution algorithm version 1.

    Algorithm:
        Monkey King Evolution version 1

    Date:
        2018

    Authors:
        Klemen Berkovič

    License:
        MIT

    Reference URL:
        https://www.sciencedirect.com/science/article/pii/S0950705116000198

    Reference paper:
        Zhenyu Meng, Jeng-Shyang Pan, Monkey King Evolution: A new memetic evolutionary algorithm and its application in vehicle fuel consumption optimization, Knowledge-Based Systems, Volume 97, 2016, Pages 144-157, ISSN 0950-7051, https://doi.org/10.1016/j.knosys.2016.01.009.

    Attributes:
        Name (List[str]): List of strings representing algorithm names.
        fluctuation_coeff (float): Scale factor for normal particles.
        population_rate (float): Percent value of now many new particle Monkey King particle creates.
        c (int): Number of new particles generated by Monkey King particle.
        fc (float): Scale factor for Monkey King particles.

    See Also:
        * :class:`niapy.algorithms.algorithm.Algorithm`

    """

    Name = ['MonkeyKingEvolutionV1', 'MKEv1']

    @staticmethod
    def info():
        r"""Get basic information of algorithm.

        Returns:
            str: Basic information.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""Zhenyu Meng, Jeng-Shyang Pan, Monkey King Evolution: A new memetic evolutionary algorithm and its application in vehicle fuel consumption optimization, Knowledge-Based Systems, Volume 97, 2016, Pages 144-157, ISSN 0950-7051, https://doi.org/10.1016/j.knosys.2016.01.009."""

    def __init__(self, population_size=40, fluctuation_coeff=0.7, population_rate=0.3, c=3, fc=0.5, *args, **kwargs):
        """Initialize MonkeyKingEvolutionV1.

        Args:
            population_size (int): Population size.
            fluctuation_coeff (float): Scale factor for normal particle.
            population_rate (float): Percent value of now many new particle Monkey King particle creates. Value in rage [0, 1].
            c (int): Number of new particles generated by Monkey King particle.
            fc (float): Scale factor for Monkey King particles.

        See Also:
            * :func:`niapy.algorithms.algorithm.Algorithm.__init__`

        """
        super().__init__(population_size,
                         individual_type=kwargs.pop('individual_type', MkeSolution),
                         initialization_function=kwargs.pop('initialization_function', default_individual_init),
                         *args, **kwargs)
        self.fluctuation_coeff = fluctuation_coeff
        self.population_rate = population_rate
        self.c = c
        self.fc = fc

    def set_parameters(self, population_size=40, fluctuation_coeff=0.7, population_rate=0.3, c=3, fc=0.5, **kwargs):
        r"""Set Monkey King Evolution v1 algorithms static parameters.

        Args:
            population_size (int): Population size.
            fluctuation_coeff (float): Scale factor for normal particle.
            population_rate (float): Percent value of now many new particle Monkey King particle creates. Value in rage [0, 1].
            c (int): Number of new particles generated by Monkey King particle.
            fc (float): Scale factor for Monkey King particles.

        See Also:
            * :func:`niapy.algorithms.algorithm.Algorithm.set_parameters`

        """
        super().set_parameters(population_size=population_size,
                               individual_type=kwargs.pop('individual_type', MkeSolution),
                               initialization_function=kwargs.pop('initialization_function', default_individual_init),
                               **kwargs)
        self.fluctuation_coeff = fluctuation_coeff
        self.population_rate = population_rate
        self.c = c
        self.fc = fc

    def get_parameters(self):
        r"""Get algorithms parameters values.

        Returns:
            Dict[str, Any]

        See Also:
            * :func:`niapy.algorithms.Algorithm.get_parameters`

        """
        d = Algorithm.get_parameters(self)
        d.update({
            'fluctuation_coeff': self.fluctuation_coeff,
            'population_rate': self.population_rate,
            'c': self.c,
            'fc': self.fc
        })
        return d

    def move_p(self, x, x_pb, x_b, task):
        r"""Move normal particle in search space.

        For moving particles algorithm uses next formula:
        :math:`\mathbf{x_{pb} - \mathit{differential_weight} \odot \mathbf{r} \odot (\mathbf{x_b} - \mathbf{x})`
        where
        :math:`\mathbf{r}` is one dimension array with `D` components. Components in this vector are in range [0, 1].

        Args:
            x (numpy.ndarray): Particle position.
            x_pb (numpy.ndarray): Particle best position.
            x_b (numpy.ndarray): Best particle position.
            task (Task): Optimization task.

        Returns:
            numpy.ndarray: Particle new position.

        """
        return x_pb + self.fluctuation_coeff * self.random(task.dimension) * (x_b - x)

    def move_mk(self, x, task):
        r"""Move Monkey King particle.

        For moving Monkey King particles algorithm uses next formula:
        :math:`\mathbf{x} + \mathit{fc} \odot \mathbf{population_rate} \odot \mathbf{x}`
        where
        :math:`\mathbf{population_rate}` is two dimensional array with shape `{c * D, D}`. Components of this array are in range [0, 1]

        Args:
            x (numpy.ndarray): Monkey King patricle position.
            task (Task): Optimization task.

        Returns:
            numpy.ndarray: New particles generated by Monkey King particle.

        """
        return x + self.fc * self.random((int(self.c * task.dimension), task.dimension)) * x

    def move_particle(self, p, p_b, task):
        r"""Move particles.

        Args:
            p (MkeSolution): Monkey particle.
            p_b (numpy.ndarray): Population best particle.
            task (Task): Optimization task.

        """
        p.x = self.move_p(p.x, p.x_pb, p_b, task)
        p.evaluate(task, rng=self.rng)

    def move_monkey_king_particle(self, p, task):
        r"""Move Monkey King Particles.

        Args:
            p (MkeSolution): Monkey King particle to apply this function on.
            task (Task): Optimization task.

        """
        p.monkey_king = False
        a = np.apply_along_axis(task.repair, 1, self.move_mk(p.x, task), self.rng)
        a_f = np.apply_along_axis(task.eval, 1, a)
        ib = np.argmin(a_f)
        p.x, p.f = a[ib], a_f[ib]

    def move_population(self, pop, xb, task):
        r"""Move population.

        Args:
            pop (numpy.ndarray[MkeSolution]): Current population.
            xb (numpy.ndarray): Current best solution.
            task (Task): Optimization task.

        Returns:
            numpy.ndarray[MkeSolution]: New particles.

        """
        for p in pop:
            if p.monkey_king:
                self.move_monkey_king_particle(p, task)
            else:
                self.move_particle(p, xb, task)
            p.update_personal_best()
        return pop

    def init_population(self, task):
        r"""Init population.

        Args:
            task (Task): Optimization task

        Returns:
            Tuple(numpy.ndarray[MkeSolution], numpy.ndarray[float], Dict[str, Any]]:
                1. Initialized solutions
                2. Fitness/function values of solution
                3. Additional arguments

        """
        pop, fpop, _ = Algorithm.init_population(self, task)
        for i in self.rng.choice(self.population_size, int(self.population_rate * len(pop)), replace=False):
            pop[i].monkey_king = True
        return pop, fpop, {}

    def run_iteration(self, task, population, population_fitness, best_x, best_fitness, **params):
        r"""Core function of Monkey King Evolution v1 algorithm.

        Args:
            task (Task): Optimization task.
            population (numpy.ndarray[MkeSolution]): Current population.
            population_fitness (numpy.ndarray[float]): Current population fitness/function values.
            best_x (numpy.ndarray): Current best solution.
            best_fitness (float): Current best solutions function/fitness value.
            **params (Dict[str, Any]): Additional arguments.

        Returns:
            Tuple(numpy.ndarray[MkeSolution], numpy.ndarray[float], Dict[str, Any]]:
                1. Initialized solutions.
                2. Fitness/function values of solution.
                3. Additional arguments.

        """
        population = self.move_population(population, best_x, task)
        for i in self.rng.choice(self.population_size, int(self.population_rate * len(population)), replace=False):
            population[i].monkey_king = True
        population_fitness = np.asarray([m.f for m in population])
        best_x, best_fitness = self.get_best(population, population_fitness, best_x, best_fitness)
        return population, population_fitness, best_x, best_fitness, {}


class MonkeyKingEvolutionV2(MonkeyKingEvolutionV1):
    r"""Implementation of monkey king evolution algorithm version 2.

    Algorithm:
        Monkey King Evolution version 2

    Date:
        2018

    Authors:
        Klemen Berkovič

    License:
        MIT

    Reference URL:
        https://www.sciencedirect.com/science/article/pii/S0950705116000198

    Reference paper:
        Zhenyu Meng, Jeng-Shyang Pan, Monkey King Evolution: A new memetic evolutionary algorithm and its application in vehicle fuel consumption optimization, Knowledge-Based Systems, Volume 97, 2016, Pages 144-157, ISSN 0950-7051, https://doi.org/10.1016/j.knosys.2016.01.009.

    Attributes:
        Name (List[str]): List of strings representing algorithm names.

    See Also:
        * :class:`niapy.algorithms.basic.mke.MonkeyKingEvolutionV1`

    """

    Name = ['MonkeyKingEvolutionV2', 'MKEv2']

    @staticmethod
    def info():
        r"""Get basic information of algorithm.

        Returns:
            str: Basic information.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""Zhenyu Meng, Jeng-Shyang Pan, Monkey King Evolution: A new memetic evolutionary algorithm and its application in vehicle fuel consumption optimization, Knowledge-Based Systems, Volume 97, 2016, Pages 144-157, ISSN 0950-7051, https://doi.org/10.1016/j.knosys.2016.01.009."""

    def move_mk(self, x, task, dx=None):
        r"""Move Monkey King particle.

        For movement of particles algorithm uses next formula:
        :math:`\mathbf{x} - \mathit{fc} \odot \mathbf{dx}`

        Args:
            x (numpy.ndarray): Particle to apply movement on.
            task (Task): Optimization task.
            dx (numpy.ndarray): Difference between to random particles in population.

        Returns:
            numpy.ndarray: Moved particles.

        """
        return x - self.fc * dx

    def move_monkey_king_particle(self, p, task, pop=None):
        r"""Move Monkey King particles.

        Args:
            p (MkeSolution): Monkey King particle to move.
            task (Task): Optimization task.
            pop (numpy.ndarray[MkeSolution]): Current population.

        """
        p.monkey_king = False
        p_b, p_f = p.x, p.f
        for _i in range(int(self.c * self.population_size)):
            r = self.rng.choice(self.population_size, 2, replace=False)
            a = task.repair(self.move_mk(p.x, task, pop[r[0]].x - pop[r[1]].x), self.rng)
            a_f = task.eval(a)
            if a_f < p_f:
                p_b, p_f = a, a_f
        p.x, p.f = p_b, p_f

    def move_population(self, pop, xb, task):
        r"""Move population.

        Args:
            pop (numpy.ndarray[MkeSolution]): Current population.
            xb (numpy.ndarray): Current best solution.
            task (Task): Optimization task.

        Returns:
            numpy.ndarray[MkeSolution]: Moved population.

        """
        for p in pop:
            if p.monkey_king:
                self.move_monkey_king_particle(p, task, pop)
            else:
                self.move_particle(p, xb, task)
            p.update_personal_best()
        return pop


class MonkeyKingEvolutionV3(MonkeyKingEvolutionV1):
    r"""Implementation of monkey king evolution algorithm version 3.

    Algorithm:
        Monkey King Evolution version 3

    Date:
        2018

    Authors:
        Klemen Berkovič

    License:
        MIT

    Reference URL:
        https://www.sciencedirect.com/science/article/pii/S0950705116000198

    Reference paper:
        Zhenyu Meng, Jeng-Shyang Pan, Monkey King Evolution: A new memetic evolutionary algorithm and its application in vehicle fuel consumption optimization, Knowledge-Based Systems, Volume 97, 2016, Pages 144-157, ISSN 0950-7051, https://doi.org/10.1016/j.knosys.2016.01.009.

    Attributes:
        Name (List[str]): List of strings that represent algorithm names.

    See Also:
        * :class:`niapy.algorithms.basic.mke.MonkeyKingEvolutionV1`

    """

    Name = ['MonkeyKingEvolutionV3', 'MKEv3']

    @staticmethod
    def info():
        r"""Get basic information of algorithm.

        Returns:
            str: Basic information.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""Zhenyu Meng, Jeng-Shyang Pan, Monkey King Evolution: A new memetic evolutionary algorithm and its application in vehicle fuel consumption optimization, Knowledge-Based Systems, Volume 97, 2016, Pages 144-157, ISSN 0950-7051, https://doi.org/10.1016/j.knosys.2016.01.009."""

    def __init__(self, *args, **kwargs):
        """Initialize MonkeyKingEvolutionV3."""
        super().__init__(individual_type=kwargs.pop('individual_type', None),
                         initialization_function=kwargs.pop('initialization_function', default_numpy_init),
                         *args, **kwargs)

    def set_parameters(self, **kwargs):
        r"""Set core parameters of MonkeyKingEvolutionV3 algorithm.

        See Also:
            * :func:`niapy.algorithms.basic.MonkeyKingEvolutionV1.set_parameters`

        """
        super().set_parameters(individual_type=kwargs.pop('individual_type', None),
                               initialization_function=kwargs.pop('initialization_function', default_numpy_init),
                               **kwargs)

    @staticmethod
    def neg(x):
        r"""Transform function.

        Args:
            x (Union[int, float]): Should be 0 or 1.

        Returns:
            float: If 0 then 1 else 0.

        """
        return 0.0 if x == 1.0 else 1.0

    def init_population(self, task):
        r"""Initialize the population.

        Args:
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray[float], Dict[str, Any]]:
                1. Initialized population.
                2. Initialized population function/fitness values.
                3. Additional arguments:
                    * k (int): Starting number of rows to include from lower triangular matrix.
                    * c (int): Constant.

        See Also:
            * :func:`niapy.algorithms.algorithm.Algorithm.init_population`

        """
        population, fitness, d = Algorithm.init_population(self, task)
        k, c = int(ceil(self.population_size / task.dimension)), int(ceil(self.c * task.dimension))
        d.update({'k': k, 'c': c})
        return population, fitness, d

    def run_iteration(self, task, population, population_fitness, best_x, best_fitness, **params):
        r"""Core function of Monkey King Evolution v3 algorithm.

        Args:
            task (Task): Optimization task.
            population (numpy.ndarray): Current population.
            population_fitness (numpy.ndarray[float]): Current population fitness/function values.
            best_x (numpy.ndarray): Current best individual.
            best_fitness (float): Current best individual function/fitness value.
            **params: Additional arguments

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray[float], Dict[str, Any]]:
                1. Initialized population.
                2. Initialized population function/fitness values.
                3. Additional arguments:
                    * k (int): Starting number of rows to include from lower triangular matrix.
                    * c (int): Constant.

        """
        k = params.pop('k')
        c = params.pop('c')

        x_gb = np.apply_along_axis(task.repair, 1,
                                   best_x + self.fc * population[self.rng.choice(len(population), c)] - population[self.rng.choice(len(population), c)],
                                   self.rng)
        x_gb_f = np.apply_along_axis(task.eval, 1, x_gb)
        best_x, best_fitness = self.get_best(x_gb, x_gb_f, best_x, best_fitness)
        m = np.ones((self.population_size, task.dimension))
        for i in range(k):
            m[i * task.dimension:(i + 1) * task.dimension] = np.tril(m[i * task.dimension:(i + 1) * task.dimension])
        for i in range(self.population_size):
            self.rng.shuffle(m[i])
        population = np.apply_along_axis(task.repair, 1, m * population + np.vectorize(self.neg)(m) * best_x, self.rng)
        population_fitness = np.apply_along_axis(task.eval, 1, population)
        best_x, best_fitness = self.get_best(population, population_fitness, best_x, best_fitness)
        iw, ib_gb = np.argmax(population_fitness), np.argmin(x_gb_f)
        if x_gb_f[ib_gb] <= population_fitness[iw]:
            population[iw], population_fitness[iw] = x_gb[ib_gb], x_gb_f[ib_gb]
        return population, population_fitness, best_x, best_fitness, {'k': k, 'c': c}



# encoding=utf8

"""Whitley function."""

import numpy as np
from niapy.problems.problem import Problem

__all__ = ['Whitley']


class Whitley(Problem):
    r"""Implementation of Whitley function.

    Date: 2018

    Authors: Grega Vrbančič and Lucija Brezočnik

    License: MIT

    Function: **Whitley function**

        :math:`f(\mathbf{x}) = \sum_{i=1}^D \sum_{j=1}^D
        \left(\frac{(100(x_i^2-x_j)^2 + (1-x_j)^2)^2}{4000} -
        \cos(100(x_i^2-x_j)^2 + (1-x_j)^2)+1\right)`

        **Input domain:**
        The function can be defined on any input domain but it is usually
        evaluated on the hypercube :math:`x_i ∈ [-10.24, 10.24]`, for all :math:`i = 1, 2,..., D`.

        **Global minimum:** :math:`f(x^*) = 0`, at :math:`x^* = (1,...,1)`

    LaTeX formats:
        Inline:
            $f(\mathbf{x}) =
            \sum_{i=1}^D \sum_{j=1}^D \left(\frac{(100(x_i^2-x_j)^2 +
            (1-x_j)^2)^2}{4000} - \cos(100(x_i^2-x_j)^2 + (1-x_j)^2)+1\right)$

        Equation:
            \begin{equation}f(\mathbf{x}) =
            \sum_{i=1}^D \sum_{j=1}^D \left(\frac{(100(x_i^2-x_j)^2 +
            (1-x_j)^2)^2}{4000} - \cos(100(x_i^2-x_j)^2 +
            (1-x_j)^2)+1\right) \end{equation}

        Domain:
            $-10.24 \leq x_i \leq 10.24$

    Reference paper:
        Jamil, M., and Yang, X. S. (2013).
        A literature survey of benchmark functions for global optimisation problems.
        International Journal of Mathematical Modelling and Numerical Optimisation,
        4(2), 150-194.

    """

    def __init__(self, dimension=4, lower=-10.24, upper=10.24, *args, **kwargs):
        r"""Initialize Whitley problem..

        Args:
            dimension (Optional[int]): Dimension of the problem.
            lower (Optional[Union[float, Iterable[float]]]): Lower bounds of the problem.
            upper (Optional[Union[float, Iterable[float]]]): Upper bounds of the problem.

        See Also:
            :func:`niapy.problems.Problem.__init__`

        """
        super().__init__(dimension, lower, upper, *args, **kwargs)

    @staticmethod
    def latex_code():
        r"""Return the latex code of the problem.

        Returns:
            str: Latex code.

        """
        return r"""$f(\mathbf{x}) =
                \sum_{i=1}^D \sum_{j=1}^D \left(\frac{(100(x_i^2-x_j)^2 +
                (1-x_j)^2)^2}{4000} - \cos(100(x_i^2-x_j)^2 + (1-x_j)^2)+1\right)$"""

    def _evaluate(self, x):
        xi = np.tile(x, (self.dimension, 1)).T
        xj = np.tile(x, (self.dimension, 1))
        tmp = 100.0 * (xi ** 2 - xj) ** 2 + (1 - xj) ** 2
        return np.sum((tmp ** 2) / 4000.0 - np.cos(tmp) + 1.0)



# encoding=utf8
import numpy as np

from niapy.algorithms.algorithm import Algorithm, Individual
from niapy.util import objects_to_array


class Fish(Individual):
    r"""Fish individual class.

    Attributes:
        weight (float): Weight of fish.
        delta_pos (float): Displacement due to individual movement.
        delta_cost (float): Cost at `delta_apos`.
        has_improved (bool): True if the fish has improved.

    See Also:
        * :class:`niapy.algorithms.algorithm.Individual`

    """

    def __init__(self, weight, **kwargs):
        r"""Initialize fish individual.

        Args:
            weight (float): Weight of fish.

        See Also:
            * :func:`niapy.algorithms.algorithm.Individual`

        """
        super().__init__(**kwargs)
        self.weight = weight
        self.delta_pos = np.nan
        self.delta_cost = np.nan
        self.has_improved = False


class FishSchoolSearch(Algorithm):
    r"""Implementation of Fish School Search algorithm.

    Algorithm:
        Fish School Search algorithm

    Date:
        2019

    Authors:
        Clodomir Santana Jr, Elliackin Figueredo, Mariana Maceds, Pedro Santos.
        Ported to niapy with small changes by Kristian Järvenpää (2018).
        Ported to niapy 2.0 by Klemen Berkovič (2019).

    License:
        MIT

    Reference paper:
        Bastos Filho, Lima Neto, Lins, D. O. Nascimento and P. Lima,
        “A novel search algorithm based on fish school behavior,”
        in 2008 IEEE International Conference on Systems, Man and Cybernetics, Oct 2008, pp. 2646–2651.

    Attributes:
        Name (List[str]): List of strings representing algorithm name.
        step_individual_init (float): Length of initial individual step.
        step_individual_final (float): Length of final individual step.
        step_volitive_init (float): Length of initial volatile step.
        step_volitive_final (float): Length of final volatile step.
        min_w (float): Minimum weight of a fish.
        w_scale (float): Maximum weight of a fish.

    See Also:
        * :class:`niapy.algorithms.algorithm.Algorithm`

    """

    Name = ['FSS', 'FishSchoolSearch']

    @staticmethod
    def info():
        r"""Get default information of algorithm.

        Returns:
            str: Basic information.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""Bastos Filho, Lima Neto, Lins, D. O. Nascimento and P. Lima,
        “A novel search algorithm based on fish school behavior,”
        in 2008 IEEE International Conference on Systems, Man and Cybernetics, Oct 2008, pp. 2646–2651."""

    def __init__(self, population_size=30, step_individual_init=0.1, step_individual_final=0.0001,
                 step_volitive_init=0.01, step_volitive_final=0.001, min_w=1.0, w_scale=500.0, *args, **kwargs):
        """Initialize FishSchoolSearch.

        Args:
            population_size (Optional[int]): Number of fishes in school.
            step_individual_init (Optional[float]): Length of initial individual step.
            step_individual_final (Optional[float]): Length of final individual step.
            step_volitive_init (Optional[float]): Length of initial volatile step.
            step_volitive_final (Optional[float]): Length of final volatile step.
            min_w (Optional[float]): Minimum weight of a fish.
            w_scale (Optional[float]): Maximum weight of a fish. Recommended value: max_iterations / 2

        See Also:
            * :func:`niapy.algorithms.Algorithm.__init__`

        """
        super().__init__(population_size, *args, **kwargs)
        self.step_individual_init = step_individual_init
        self.step_individual_final = step_individual_final
        self.step_volitive_init = step_volitive_init
        self.step_volitive_final = step_volitive_final
        self.min_w = min_w
        self.w_scale = w_scale

    def set_parameters(self, population_size=30, step_individual_init=0.1, step_individual_final=0.0001,
                       step_volitive_init=0.01, step_volitive_final=0.001, min_w=1.0, w_scale=5000.0, **kwargs):
        r"""Set core arguments of FishSchoolSearch algorithm.

        Args:
            population_size (Optional[int]): Number of fishes in school.
            step_individual_init (Optional[float]): Length of initial individual step.
            step_individual_final (Optional[float]): Length of final individual step.
            step_volitive_init (Optional[float]): Length of initial volatile step.
            step_volitive_final (Optional[float]): Length of final volatile step.
            min_w (Optional[float]): Minimum weight of a fish.
            w_scale (Optional[float]): Maximum weight of a fish. Recommended value: max_iterations / 2

        See Also:
            * :func:`niapy.algorithms.Algorithm.set_parameters`

        """
        super().set_parameters(population_size=population_size, **kwargs)
        self.step_individual_init = step_individual_init
        self.step_individual_final = step_individual_final
        self.step_volitive_init = step_volitive_init
        self.step_volitive_final = step_volitive_final
        self.min_w = min_w
        self.w_scale = w_scale

    def get_parameters(self):
        r"""Get algorithm parameters.

        Returns:
            Dict[str, Any]: Algorithm parameters.

        See Also:
            * :func:`niapy.algorithms.Algorithm.set_parameters`

        """
        d = super().get_parameters()
        d.update({
            'step_individual_init': self.step_individual_init,
            'step_individual_final': self.step_individual_final,
            'step_volitive_init': self.step_volitive_init,
            'step_volitive_final': self.step_volitive_final,
            'min_w': self.min_w,
            'w_scale': self.w_scale
        })
        return d

    def init_school(self, task):
        """Initialize fish school with uniform distribution."""
        step_individual = self.step_individual_init * task.range
        step_volitive = self.step_volitive_init * task.range
        school = [Fish(weight=self.w_scale / 2.0, task=task, e=True, rng=self.rng) for _ in range(self.population_size)]
        school_weight = self.population_size * self.w_scale / 2.0
        return step_individual, step_volitive, school_weight, objects_to_array(school)

    def update_steps(self, task):
        r"""Update step length for individual and volatile steps.

        Args:
            task (Task): Optimization task

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray]:
                1. New individual step.
                2. New volitive step.

        """
        step_individual = np.full(task.dimension, self.step_individual_init - (task.iters + 1) * (
                    self.step_individual_init - self.step_individual_final) / task.max_iters)
        step_volitive = np.full(task.dimension, self.step_volitive_init - (task.iters + 1) * (
                    self.step_volitive_init - self.step_volitive_final) / task.max_iters)
        return step_individual, step_volitive

    def feeding(self, school):
        r"""Feed all fishes.

        Args:
            school (numpy.ndarray): Current school fish population.

        Returns:
            numpy.ndarray: New school fish population.

        """
        max_delta_cost = max(fish.delta_cost for fish in school)
        for fish in school:
            if max_delta_cost:
                fish.weight = fish.weight + (fish.delta_cost / max_delta_cost)
            fish.weight = np.clip(fish.weight, self.min_w, self.w_scale)
        return school

    def individual_movement(self, school, step_individual, xb, fxb, task):
        r"""Perform individual movement for each fish.

        Args:
            school (numpy.ndarray): School fish population.
            step_individual (numpy.ndarray): Current individual step.
            xb (numpy.ndarray): Global best solution.
            fxb (float): Global best solutions fitness/objective value.
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, float]:
                1. New school of fishes.
                2. New global best position.
                3. New global best fitness.

        """
        for fish in school:
            new_pos = task.repair(fish.x + (step_individual * self.uniform(-1, 1, task.dimension)), rng=self.rng)
            cost = task.eval(new_pos)
            if cost < fish.f:
                xb, fxb = self.get_best(new_pos, cost, xb, fxb)
                fish.delta_cost = abs(cost - fish.f)
                fish.f = cost
                fish.delta_pos = new_pos - fish.x
                fish.x = new_pos
            else:
                fish.delta_pos = np.zeros(task.dimension)
                fish.delta_cost = 0
        return school, xb, fxb

    def collective_instinctive_movement(self, school, task):
        r"""Perform collective instinctive movement.

        Args:
            school (numpy.ndarray): Current population.
            task (Task): Optimization task.

        Returns:
            numpy.ndarray: New population

        """
        cost_eval_enhanced = sum((fish.delta_cost * fish.delta_pos for fish in school), start=np.zeros(task.dimension))
        density = sum(f.delta_cost for f in school)
        if density != 0:
            cost_eval_enhanced /= density
        for fish in school:
            fish.x = task.repair(fish.x + cost_eval_enhanced, rng=self.rng)
        return school

    def collective_volitive_movement(self, school, step_volitive, school_weight, xb, fxb, task):
        r"""Perform collective volitive movement.

        Args:
            school (numpy.ndarray):
            step_volitive :
            school_weight:
            xb (numpy.ndarray): Global best solution.
            fxb (float): Global best solutions fitness/objective value.
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, float]:
                1. New population.
                2. New global best individual.
                3. New global best fitness.

        """
        prev_weight_school = school_weight
        school_weight = sum(fish.weight for fish in school)

        barycenter = sum((fish.x * fish.weight for fish in school), start=np.zeros(task.dimension))
        barycenter /= sum(fish.weight for fish in school)
        for fish in school:
            if school_weight > prev_weight_school:
                fish.x -= (fish.x - barycenter) * step_volitive * self.uniform(0, 1, task.dimension)
            else:
                fish.x += (fish.x - barycenter) * step_volitive * self.uniform(0, 1, task.dimension)
            fish.evaluate(task, rng=self.rng)
            xb, fxb = self.get_best(fish.x, fish.f, xb, fxb)
        return school, xb, fxb

    def init_population(self, task):
        r"""Initialize the school.

        Args:
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, dict]:
                1. Population.
                2. Population fitness.
                3. Additional arguments:
                    * step_individual (float): Current individual step.
                    * step_volitive (float): Current volitive step.
                    * school_weight (float): Current school weight.

        """
        step_individual, step_volitive, school_weight, school = self.init_school(task)
        return school, np.asarray([f.f for f in school]), {'step_individual': step_individual,
                                                           'step_volitive': step_volitive,
                                                           'school_weight': school_weight}

    def run_iteration(self, task, population, population_fitness, best_x, best_fitness, **params):
        r"""Core function of algorithm.

        Args:
            task (Task): Optimization task.
            population (numpy.ndarray): Current population.
            population_fitness (numpy.ndarray): Current population fitness.
            best_x (numpy.ndarray): Current global best individual.
            best_fitness (float): Current global best fitness.
            **params: Additional parameters.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, float, dict]:
                1. New Population.
                2. New Population fitness.
                3. New global best individual.
                4. New global best fitness.
                5. Additional parameters:
                    * step_individual (float): Current individual step.
                    * step_volitive (float): Current volitive step.
                    * school_weight (float): Current school weight.

        """
        step_individual = params.pop('step_individual')
        step_volitive = params.pop('step_volitive')
        school_weight = params.pop('school_weight')

        population, best_x, best_fitness = self.individual_movement(population, step_individual, best_x, best_fitness, task)
        population = self.feeding(population)
        population = self.collective_instinctive_movement(population, task)
        population, best_x, best_fitness = self.collective_volitive_movement(population, step_volitive, school_weight,
                                                                             best_x, best_fitness, task)
        step_individual, step_volitive = self.update_steps(task)
        return population, np.asarray([f.f for f in population]), best_x, best_fitness, {'step_individual': step_individual,
                                                                                         'step_volitive': step_volitive,
                                                                                         'school_weight': school_weight
                                                                                         }



# encoding=utf8
import logging
import operator

import numpy as np

from niapy.algorithms.algorithm import Algorithm

logging.basicConfig()
logger = logging.getLogger('niapy.algorithms.other')
logger.setLevel('INFO')

__all__ = ['MultipleTrajectorySearch', 'MultipleTrajectorySearchV1', 'mts_ls1', 'mts_ls1v1', 'mts_ls2', 'mts_ls3',
           'mts_ls3v1']


def mts_ls1(current_x, current_fitness, best_x, best_fitness, improve, search_range, task, rng, bonus1=10, bonus2=1,
            sr_fix=0.4, **_kwargs):
    r"""Multiple trajectory local search one.

    Args:
        current_x (numpy.ndarray): Current solution.
        current_fitness (float): Current solutions fitness/function value.
        best_x (numpy.ndarray): Global best solution.
        best_fitness (float): Global best solutions fitness/function value.
        improve (bool): Has the solution been improved.
        search_range (numpy.ndarray): Search range.
        task (Task): Optimization task.
        rng (numpy.random.Generator): Random number generator.
        bonus1 (int): Bonus reward for improving global best solution.
        bonus2 (int): Bonus reward for improving solution.
        sr_fix (numpy.ndarray): Fix when search range is to small.

    Returns:
        Tuple[numpy.ndarray, float, numpy.ndarray, float, bool, numpy.ndarray]:
            1. New solution.
            2. New solutions fitness/function value.
            3. Global best if found else old global best.
            4. Global bests function/fitness value.
            5. If solution has improved.
            6. Search range.

    """
    if not improve:
        search_range /= 2
        i_fix = np.argwhere(search_range < 1e-15)
        search_range[i_fix] = task.range[i_fix] * sr_fix
    improve = False
    grade = 0.0
    for i in range(len(current_x)):
        x_old = current_x[i]
        current_x[i] = x_old - search_range[i]
        current_x = task.repair(current_x, rng)
        new_fitness = task.eval(current_x)
        if new_fitness < best_fitness:
            grade = grade + bonus1
            best_x = current_x.copy()
            best_fitness = new_fitness
        if new_fitness == current_fitness:
            current_x[i] = x_old
        elif new_fitness > current_fitness:
            current_x[i] = x_old + 0.5 * search_range[i]
            current_x = task.repair(current_x, rng)
            new_fitness = task.eval(current_x)
            if new_fitness < best_fitness:
                grade = grade + bonus1
                best_x = current_x.copy()
                best_fitness = new_fitness
            if new_fitness >= current_fitness:
                current_x[i] = x_old
            else:
                grade = grade + bonus2
                improve = True
                current_fitness = new_fitness
        else:
            grade = grade + bonus2
            improve = True
            current_fitness = new_fitness
    return current_x, current_fitness, best_x, best_fitness, improve, grade, search_range


def mts_ls1v1(current_x, current_fitness, best_x, best_fitness, improve, search_range, task, rng, bonus1=10, bonus2=1,
              sr_fix=0.4, **_kwargs):
    r"""Multiple trajectory local search one version two.

    Args:
        current_x (numpy.ndarray): Current solution.
        current_fitness (float): Current solutions fitness/function value.
        best_x (numpy.ndarray): Global best solution.
        best_fitness (float): Global best solutions fitness/function value.
        improve (bool): Has the solution been improved.
        search_range (numpy.ndarray): Search range.
        task (Task): Optimization task.
        rng (numpy.random.Generator): Random number generator.
        bonus1 (int): Bonus reward for improving global best solution.
        bonus2 (int): Bonus reward for improving solution.
        sr_fix (numpy.ndarray): Fix when search range is to small.

    Returns:
        Tuple[numpy.ndarray, float, numpy.ndarray, float, bool, numpy.ndarray]:
            1. New solution.
            2. New solutions fitness/function value.
            3. Global best if found else old global best.
            4. Global bests function/fitness value.
            5. If solution has improved.
            6. Search range.

    """
    if not improve:
        search_range /= 2
        i_fix = np.argwhere(search_range < 1e-15)
        search_range[i_fix] = task.range[i_fix] * sr_fix
    improve, d, grade = False, rng.uniform(-1, 1, task.dimension), 0.0
    for i in range(len(current_x)):
        x_old = current_x[i]
        current_x[i] = x_old - search_range[i] * d[i]
        current_x = task.repair(current_x, rng)
        new_fitness = task.eval(current_x)
        if new_fitness < best_fitness:
            grade, best_x, best_fitness = grade + bonus1, current_x.copy(), new_fitness
        elif new_fitness == current_fitness:
            current_x[i] = x_old
        elif new_fitness > current_fitness:
            current_x[i] = x_old + 0.5 * search_range[i]
            current_x = task.repair(current_x, rng)
            new_fitness = task.eval(current_x)
            if new_fitness < best_fitness:
                grade, best_x, best_fitness = grade + bonus1, current_x.copy(), new_fitness
            elif new_fitness >= current_fitness:
                current_x[i] = x_old
            else:
                grade, improve, current_fitness = grade + bonus2, True, new_fitness
        else:
            grade, improve, current_fitness = grade + bonus2, True, new_fitness
    return current_x, current_fitness, best_x, best_fitness, improve, grade, search_range


def move_x(x, r, d, search_range, op):
    r"""Move solution to other position based on operator.

    Args:
        x (numpy.ndarray): Solution to move.
        r (int): Random number.
        d (float): Scale factor.
        search_range (numpy.ndarray): Search range.
        op (Callable): Operator to use.

    Returns:
        numpy.ndarray: Moved solution based on operator.

    """
    return op(x, search_range * d) if r == 0 else x


def mts_ls2(current_x, current_fitness, best_x, best_fitness, improve, search_range, task, rng, bonus1=10, bonus2=1,
            sr_fix=0.4, **_kwargs):
    r"""Multiple trajectory local search two.

    Args:
        current_x (numpy.ndarray): Current solution.
        current_fitness (float): Current solutions fitness/function value.
        best_x (numpy.ndarray): Global best solution.
        best_fitness (float): Global best solutions fitness/function value.
        improve (bool): Has the solution been improved.
        search_range (numpy.ndarray): Search range.
        task (Task): Optimization task.
        rng (numpy.random.Generator): Random number generator.
        bonus1 (int): Bonus reward for improving global best solution.
        bonus2 (int): Bonus reward for improving solution.
        sr_fix (numpy.ndarray): Fix when search range is to small.

    Returns:
        Tuple[numpy.ndarray, float, numpy.ndarray, float, bool, numpy.ndarray]:
            1. New solution.
            2. New solutions fitness/function value.
            3. Global best if found else old global best.
            4. Global bests function/fitness value.
            5. If solution has improved.
            6. Search range.

    See Also:
        * :func:`niapy.algorithms.other.move_x`

    """
    if not improve:
        search_range /= 2
        i_fix = np.argwhere(search_range < 1e-15)
        search_range[i_fix] = task.range[i_fix] * sr_fix
    improve, grade = False, 0.0
    for _ in range(len(current_x)):
        d = -1 + rng.random(len(current_x)) * 2
        r = rng.choice([0, 1, 2, 3], len(current_x))
        new_x = task.repair(np.vectorize(move_x)(current_x, r, d, search_range, operator.sub), rng)
        new_fitness = task.eval(new_x)
        if new_fitness < best_fitness:
            grade, best_x, best_fitness = grade + bonus1, new_x.copy(), new_fitness
        elif new_fitness != current_fitness:
            if new_fitness > current_fitness:
                new_x = task.repair(np.vectorize(move_x)(current_x, r, d, search_range, operator.add), rng)
                new_fitness = task.eval(new_x)
                if new_fitness < best_fitness:
                    grade, best_x, best_fitness = grade + bonus1, new_x.copy(), new_fitness
                elif new_fitness < current_fitness:
                    grade, current_x, current_fitness, improve = grade + bonus2, new_x.copy(), new_fitness, True
            else:
                grade, current_x, current_fitness, improve = grade + bonus2, new_x.copy(), new_fitness, True
    return current_x, current_fitness, best_x, best_fitness, improve, grade, search_range


def mts_ls3(current_x, current_fitness, best_x, best_fitness, improve, search_range, task, rng, bonus1=10, bonus2=1,
            **_kwargs):
    r"""Multiple trajectory local search three.

    Args:
        current_x (numpy.ndarray): Current solution.
        current_fitness (float): Current solutions fitness/function value.
        best_x (numpy.ndarray): Global best solution.
        best_fitness (float): Global best solutions fitness/function value.
        improve (bool): Has the solution been improved.
        search_range (numpy.ndarray): Search range.
        task (Task): Optimization task.
        rng (numpy.random.Generator): Random number generator.
        bonus1 (int): Bonus reward for improving global best solution.
        bonus2 (int): Bonus reward for improving solution.

    Returns:
        Tuple[numpy.ndarray, float, numpy.ndarray, float, bool, numpy.ndarray]:
            1. New solution.
            2. New solutions fitness/function value.
            3. Global best if found else old global best.
            4. Global bests function/fitness value.
            5. If solution has improved.
            6. Search range.

    """
    x_new, grade = np.copy(current_x), 0.0
    for i in range(len(current_x)):
        x1, x2, x3 = np.copy(x_new), np.copy(x_new), np.copy(x_new)
        x1[i], x2[i], x3[i] = x1[i] + 0.1, x2[i] - 0.1, x3[i] + 0.2
        x1, x2, x3 = task.repair(x1, rng), task.repair(x2, rng), task.repair(x3, rng)
        x1_fit, x2_fit, x3_fit = task.eval(x1), task.eval(x2), task.eval(x3)
        if x1_fit < best_fitness:
            grade, best_x, best_fitness, improve = grade + bonus1, x1.copy(), x1_fit, True
        if x2_fit < best_fitness:
            grade, best_x, best_fitness, improve = grade + bonus1, x2.copy(), x2_fit, True
        if x3_fit < best_fitness:
            grade, best_x, best_fitness, improve = grade + bonus1, x3.copy(), x3_fit, True
        d1, d2, d3 = current_fitness - x1_fit if np.abs(x1_fit) != np.inf else 0, current_fitness - x2_fit if np.abs(
            x2_fit) != np.inf else 0, current_fitness - x3_fit if np.abs(x3_fit) != np.inf else 0
        if d1 > 0:
            grade, improve = grade + bonus2, True
        if d2 > 0:
            grade, improve = grade + bonus2, True
        if d3 > 0:
            grade, improve = grade + bonus2, True
        a, b, c = 0.4 + rng.random() * 0.1, 0.1 + rng.random() * 0.2, rng.random()
        x_new[i] += a * (d1 - d2) + b * (d3 - 2 * d1) + c
        x_new = task.repair(x_new, rng)
        x_new_fitness = task.eval(x_new)
        if x_new_fitness < current_fitness:
            if x_new_fitness < best_fitness:
                best_x, best_fitness, grade = x_new.copy(), x_new_fitness, grade + bonus1
            else:
                grade += bonus2
            current_x, current_fitness, improve = x_new, x_new_fitness, True
    return current_x, current_fitness, best_x, best_fitness, improve, grade, search_range


def mts_ls3v1(current_x, current_fitness, best_x, best_fitness, improve, search_range, task, rng, bonus1=10, bonus2=1,
              phi=3, **_kwargs):
    r"""Multiple trajectory local search three version one.

    Args:
        current_x (numpy.ndarray): Current solution.
        current_fitness (float): Current solutions fitness/function value.
        best_x (numpy.ndarray): Global best solution.
        best_fitness (float): Global best solutions fitness/function value.
        improve (bool): Has the solution been improved.
        search_range (numpy.ndarray): Search range.
        task (Task): Optimization task.
        rng (numpy.random.Generator): Random number generator.
        phi (int): Number of new generated positions.
        bonus1 (int): Bonus reward for improving global best solution.
        bonus2 (int): Bonus reward for improving solution.

    Returns:
        Tuple[numpy.ndarray, float, numpy.ndarray, float, bool, numpy.ndarray]:
            1. New solution.
            2. New solutions fitness/function value.
            3. Global best if found else old global best.
            4. Global bests function/fitness value.
            5. If solution has improved.
            6. Search range.

    """
    grade, disp = 0.0, task.range / 10
    while True in (disp > 1e-3):
        new_x = np.apply_along_axis(task.repair, 1, np.asarray(
            [rng.permutation(current_x) + disp * rng.uniform(-1, 1, len(current_x)) for _ in range(phi)]), rng)
        new_fitness = np.apply_along_axis(task.eval, 1, new_x)
        i_better, i_better_best = np.argwhere(new_fitness < current_fitness), np.argwhere(new_fitness < best_fitness)
        grade += len(i_better_best) * bonus1 + (len(i_better) - len(i_better_best)) * bonus2
        if len(new_fitness[i_better_best]) > 0:
            ib, improve = np.argmin(new_fitness[i_better_best]), True
            best_x, best_fitness, current_x, current_fitness = new_x[i_better_best][ib][0].copy(), new_fitness[i_better_best][ib][0], new_x[i_better_best][ib][
                0].copy(), new_fitness[i_better_best][ib][0]
        elif len(new_fitness[i_better]) > 0:
            ib, improve = np.argmin(new_fitness[i_better]), True
            current_x, current_fitness = new_x[i_better][ib][0].copy(), new_fitness[i_better][ib][0]
        su, sl = np.fmin(task.upper, current_x + 2 * disp), np.fmax(task.lower, current_x - 2 * disp)
        disp = (su - sl) / 10
    return current_x, current_fitness, best_x, best_fitness, improve, grade, search_range


class MultipleTrajectorySearch(Algorithm):
    r"""Implementation of Multiple trajectory search.

    Algorithm:
        Multiple trajectory search

    Date:
        2018

    Authors:
        Klemen Berkovič

    License:
        MIT

    Reference URL:
        https://ieeexplore.ieee.org/document/4631210/

    Reference paper:
        Lin-Yu Tseng and Chun Chen, "Multiple trajectory search for Large Scale Global Optimization," 2008 IEEE Congress on Evolutionary Computation (IEEE World Congress on Computational Intelligence), Hong Kong, 2008, pp. 3052-3059. doi: 10.1109/CEC.2008.4631210

    Attributes:
        Name (List[Str]): List of strings representing algorithm name.
        local_searches (Iterable[Callable[[numpy.ndarray, float, numpy.ndarray, float, bool, numpy.ndarray, Task, Dict[str, Any]], Tuple[numpy.ndarray, float, numpy.ndarray, float, bool, int, numpy.ndarray]]]): Local searches to use.
        bonus1 (int): Bonus for improving global best solution.
        bonus2 (int): Bonus for improving solution.
        num_tests (int): Number of test runs on local search algorithms.
        num_searches (int): Number of local search algorithm runs.
        num_searches_best (int): Number of locals search algorithm runs on best solution.
        num_enabled (int): Number of best solution for testing.

    See Also:
        * :class:`niapy.algorithms.Algorithm`

    """

    Name = ['MultipleTrajectorySearch', 'MTS']

    @staticmethod
    def info():
        r"""Get basic information of algorithm.

        Returns:
            str: Basic information of algorithm.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""Lin-Yu Tseng and Chun Chen, "Multiple trajectory search for Large Scale Global Optimization," 2008 IEEE Congress on Evolutionary Computation (IEEE World Congress on Computational Intelligence), Hong Kong, 2008, pp. 3052-3059. doi: 10.1109/CEC.2008.4631210"""

    def __init__(self, population_size=40, num_tests=5, num_searches=5, num_searches_best=5, num_enabled=17, bonus1=10,
                 bonus2=1, local_searches=(mts_ls1, mts_ls2, mts_ls3), *args, **kwargs):
        """Initialize MultipleTrajectorySearch.

        Args:
            population_size (int): Number of individuals in population.
            num_tests (int): Number of test runs on local search algorithms.
            num_searches (int): Number of local search algorithm runs.
            num_searches_best (int): Number of locals search algorithm runs on best solution.
            num_enabled (int): Number of best solution for testing.
            bonus1 (int): Bonus for improving global best solution.
            bonus2 (int): Bonus for improving self.
            local_searches (Iterable[Callable[[numpy.ndarray, float, numpy.ndarray, float, bool, numpy.ndarray, Task, Dict[str, Any]], Tuple[numpy.ndarray, float, numpy.ndarray, float, bool, int, numpy.ndarray]]]): Local searches to use.

        See Also:
            * :func:`niapy.algorithms.Algorithm.__init__`

        """
        super().__init__(population_size, *args, **kwargs)
        self.num_tests = num_tests
        self.num_searches = num_searches
        self.num_searches_best = num_searches_best
        self.num_enabled = num_enabled
        self.bonus1 = bonus1
        self.bonus2 = bonus2
        self.local_searches = local_searches

    def set_parameters(self, population_size=40, num_tests=5, num_searches=5, num_searches_best=5, num_enabled=17,
                       bonus1=10, bonus2=1, local_searches=(mts_ls1, mts_ls2, mts_ls3), **kwargs):
        r"""Set the arguments of the algorithm.

        Args:
            population_size (int): Number of individuals in population.
            num_tests (int): Number of test runs on local search algorithms.
            num_searches (int): Number of local search algorithm runs.
            num_searches_best (int): Number of locals search algorithm runs on best solution.
            num_enabled (int): Number of best solution for testing.
            bonus1 (int): Bonus for improving global best solution.
            bonus2 (int): Bonus for improving self.
            local_searches (Iterable[Callable[[numpy.ndarray, float, numpy.ndarray, float, bool, numpy.ndarray, Task, Dict[str, Any]], Tuple[numpy.ndarray, float, numpy.ndarray, float, bool, int, numpy.ndarray]]]): Local searches to use.

        See Also:
            * :func:`niapy.algorithms.Algorithm.set_parameters`

        """
        super().set_parameters(population_size=kwargs.pop('population_size', population_size), **kwargs)
        self.num_tests = num_tests
        self.num_searches = num_searches
        self.num_searches_best = num_searches_best
        self.num_enabled = num_enabled
        self.bonus1 = bonus1
        self.bonus2 = bonus2
        self.local_searches = local_searches

    def get_parameters(self):
        r"""Get parameters values for the algorithm.

        Returns:
            Dict[str, Any]: Algorithm parameters.

        """
        d = Algorithm.get_parameters(self)
        d.update({
            'M': d.pop('population_size', self.population_size),
            'num_tests': self.num_tests,
            'num_searches': self.num_searches,
            'num_searches_best': self.num_searches_best,
            'bonus1': self.bonus1,
            'bonus2': self.bonus2,
            'num_enabled': self.num_enabled,
            'local_searches': self.local_searches
        })
        return d

    def grading_run(self, x, x_f, xb, fxb, improve, search_range, task):
        r"""Run local search for getting scores of local searches.

        Args:
            x (numpy.ndarray): Solution for grading.
            x_f (float): Solutions fitness/function value.
            xb (numpy.ndarray): Global best solution.
            fxb (float): Global best solutions function/fitness value.
            improve (bool): Info if solution has improved.
            search_range (numpy.ndarray): Search range.
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray, float, numpy.ndarray, float]:
                1. New solution.
                2. New solutions function/fitness value.
                3. Global best solution.
                4. Global best solutions fitness/function value.

        """
        ls_grades, new_x = np.zeros(3), [[x, x_f]] * len(self.local_searches)
        k = None
        for k in range(len(self.local_searches)):
            for _ in range(self.num_tests):
                new_x[k][0], new_x[k][1], xb, fxb, improve, g, search_range = self.local_searches[k](new_x[k][0], new_x[k][1], xb, fxb, improve, search_range,
                                                                                                     task, BONUS1=self.bonus1, BONUS2=self.bonus2,
                                                                                                     rng=self.rng)
                ls_grades[k] += g
        xn, xn_f = min(new_x, key=lambda val: val[1])
        return xn, xn_f, xb, fxb, k

    def run_local_search(self, k, x, x_f, xb, fxb, improve, search_range, g, task):
        r"""Run a selected local search.

        Args:
            k (int): Index of local search.
            x (numpy.ndarray): Current solution.
            x_f (float): Current solutions function/fitness value.
            xb (numpy.ndarray): Global best solution.
            fxb (float): Global best solutions fitness/function value.
            improve (bool): If the solution has improved.
            search_range (numpy.ndarray): Search range.
            g (int): Grade.
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray, float, numpy.ndarray, float, bool, numpy.ndarray, int]:
                1. New best solution found.
                2. New best solutions found function/fitness value.
                3. Global best solution.
                4. Global best solutions function/fitness value.
                5. If the solution has improved.
                6. Grade of local search run.

        """
        for _ in range(self.num_searches):
            x, x_f, xb, fxb, improve, grade, search_range = self.local_searches[k](x, x_f, xb, fxb, improve, search_range, task, bonus1=self.bonus1,
                                                                                   bonus2=self.bonus2, rng=self.rng)
            g += grade
        return x, x_f, xb, fxb, improve, search_range, g

    def init_population(self, task):
        r"""Initialize starting population.

        Args:
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, Dict[str, Any]]:
                1. Initialized population.
                2. Initialized populations function/fitness value.
                3. Additional arguments:
                    * enable (numpy.ndarray): If solution/individual is enabled.
                    * improve (numpy.ndarray): If solution/individual is improved.
                    * search_range (numpy.ndarray): Search range.
                    * grades (numpy.ndarray): Grade of solution/individual.

        """
        population, fitness, d = Algorithm.init_population(self, task)
        enable = np.full(self.population_size, True)
        improve = np.full(self.population_size, True)
        search_range = np.full((self.population_size, task.dimension), task.range / 2)
        grades = np.zeros(self.population_size)
        d.update({
            'enable': enable,
            'improve': improve,
            'search_range': search_range,
            'grades': grades
        })
        return population, fitness, d

    def run_iteration(self, task, population, population_fitness, best_x, best_fitness, **params):
        r"""Core function of MultipleTrajectorySearch algorithm.

        Args:
            task (Task): Optimization task.
            population (numpy.ndarray): Current population of individuals.
            population_fitness (numpy.ndarray): Current individuals function/fitness values.
            best_x (numpy.ndarray): Global best individual.
            best_fitness (float): Global best individual function/fitness value.
            **params (Dict[str, Any]): Additional arguments.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, float, Dict[str, Any]]:
                1. Initialized population.
                2. Initialized populations function/fitness value.
                3. New global best solution.
                4. New global best solutions fitness/objective value.
                5. Additional arguments:
                    * enable (numpy.ndarray): If solution/individual is enabled.
                    * improve (numpy.ndarray): If solution/individual is improved.
                    * search_range (numpy.ndarray): Search range.
                    * grades (numpy.ndarray): Grade of solution/individual.

        """
        enable = params.pop('enable')
        improve = params.pop('improve')
        search_range = params.pop('search_range')
        grades = params.pop('grades')

        for i in range(len(population)):
            if not enable[i]:
                continue
            enable[i], grades[i] = False, 0
            population[i], population_fitness[i], best_x, best_fitness, k = self.grading_run(population[i], population_fitness[i], best_x, best_fitness, improve[i], search_range[i], task)
            population[i], population_fitness[i], best_x, best_fitness, improve[i], search_range[i], grades[i] = self.run_local_search(k, population[i], population_fitness[i], best_x, best_fitness, improve[i],
                                                                                                                                       search_range[i], grades[i], task)
        for _ in range(self.num_searches_best):
            _, _, best_x, best_fitness, _, _, _ = mts_ls1(best_x, best_fitness, best_x, best_fitness, False, task.range.copy() / 10, task,
                                                          rng=self.rng)
        enable[np.argsort(grades)[:self.num_enabled]] = True
        return population, population_fitness, best_x, best_fitness, {'enable': enable, 'improve': improve, 'search_range': search_range, 'grades': grades}


class MultipleTrajectorySearchV1(MultipleTrajectorySearch):
    r"""Implementation of Multiple trajectory search.

    Algorithm:
        Multiple trajectory search

    Date:
        2018

    Authors:
        Klemen Berkovič

    License:
        MIT

    Reference URL:
        https://ieeexplore.ieee.org/document/4983179/

    Reference paper:
        Tseng, Lin-Yu, and Chun Chen. "Multiple trajectory search for unconstrained/constrained multi-objective optimization." Evolutionary Computation, 2009. CEC'09. IEEE Congress on. IEEE, 2009.

    Attributes:
        Name (List[str]): List of strings representing algorithm name.

    See Also:
        * :class:`niapy.algorithms.other.MultipleTrajectorySearch``

    """

    Name = ['MultipleTrajectorySearchV1', 'MTSv1']

    @staticmethod
    def info():
        r"""Get basic information of algorithm.

        Returns:
            str: Basic information of algorithm.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""Tseng, Lin-Yu, and Chun Chen. "Multiple trajectory search for unconstrained/constrained multi-objective optimization." Evolutionary Computation, 2009. CEC'09. IEEE Congress on. IEEE, 2009."""

    def __init__(self, population_size=40, num_tests=5, num_searches=5, num_enabled=17, bonus1=10, bonus2=1, *args,
                 **kwargs):
        """Initialize MultipleTrajectorySearchV1.

        Args:
            population_size (int): Number of individuals in population.
            num_tests (int): Number of test runs on local search algorithms.
            num_searches (int): Number of local search algorithm runs.
            num_enabled (int): Number of best solution for testing.
            bonus1 (int): Bonus for improving global best solution.
            bonus2 (int): Bonus for improving self.

        See Also:
            * :func:`niapy.algorithms.other.MultipleTrajectorySearch.__init__`

        """
        kwargs.pop('num_searches_best', None)
        kwargs.pop('local_searches', None)
        super().__init__(population_size, num_tests, num_searches, 0, num_enabled, bonus1, bonus2,
                         local_searches=(mts_ls1v1, mts_ls2), *args, **kwargs)

    def set_parameters(self, population_size=40, num_tests=5, num_searches=5, num_enabled=17, bonus1=10, bonus2=1,
                       **kwargs):
        r"""Set core parameters of MultipleTrajectorySearchV1 algorithm.

        Args:
            population_size (int): Number of individuals in population.
            num_tests (int): Number of test runs on local search algorithms.
            num_searches (int): Number of local search algorithm runs.
            num_enabled (int): Number of best solution for testing.
            bonus1 (int): Bonus for improving global best solution.
            bonus2 (int): Bonus for improving self.

        See Also:
            * :func:`niapy.algorithms.other.MultipleTrajectorySearch.set_parameters`

        """
        kwargs.pop('num_searches_best', None)
        super().set_parameters(num_searches_best=0, local_searches=(mts_ls1v1, mts_ls2), **kwargs)



# encoding=utf8
"""Implementation of modified nature-inspired algorithms."""

from niapy.algorithms.modified.hba import HybridBatAlgorithm
from niapy.algorithms.modified.hde import (
    DifferentialEvolutionMTS,
    DifferentialEvolutionMTSv1,
    DynNpDifferentialEvolutionMTS,
    DynNpDifferentialEvolutionMTSv1,
    MultiStrategyDifferentialEvolutionMTS,
    DynNpMultiStrategyDifferentialEvolutionMTS,
    DynNpMultiStrategyDifferentialEvolutionMTSv1,
    MultiStrategyDifferentialEvolutionMTSv1
)
from niapy.algorithms.modified.hsaba import HybridSelfAdaptiveBatAlgorithm
from niapy.algorithms.modified.jde import (
    SelfAdaptiveDifferentialEvolution,
    MultiStrategySelfAdaptiveDifferentialEvolution,
)
from niapy.algorithms.modified.plba import ParameterFreeBatAlgorithm
from niapy.algorithms.modified.saba import (
    AdaptiveBatAlgorithm,
    SelfAdaptiveBatAlgorithm
)

__all__ = [
    'HybridBatAlgorithm',
    'DifferentialEvolutionMTS',
    'DifferentialEvolutionMTSv1',
    'DynNpDifferentialEvolutionMTS',
    'DynNpDifferentialEvolutionMTSv1',
    'MultiStrategyDifferentialEvolutionMTS',
    'MultiStrategyDifferentialEvolutionMTSv1',
    'DynNpMultiStrategyDifferentialEvolutionMTS',
    'DynNpMultiStrategyDifferentialEvolutionMTSv1',
    'SelfAdaptiveDifferentialEvolution',
    'MultiStrategySelfAdaptiveDifferentialEvolution',
    'AdaptiveBatAlgorithm',
    'SelfAdaptiveBatAlgorithm',
    'HybridSelfAdaptiveBatAlgorithm',
    'ParameterFreeBatAlgorithm'
]



# encoding=utf8
import copy
import logging
import numpy
import numpy as np

from niapy.algorithms.algorithm import Algorithm, Individual, default_individual_init
from niapy.util import objects_to_array

logging.basicConfig()
logger = logging.getLogger('niapy.algorithms.basic')
logger.setLevel('INFO')

__all__ = ['LionOptimizationAlgorithm']


class Lion(Individual):
    r"""Implementation of population individual that is a lion for Lion Optimization Algorithm.

    Algorithm:
        Lion Optimization Algorithm

    Date:
        2021

    Authors:
        Aljoša Mesarec

    License:
        MIT

    Attributes:
        gender (string): Lion gender.
        has_pride (bool): Lion has a pride.
        pride (int): Lion's pride id.
        hunting_group (int): Lion's hunting group.
        current_x (numpy.ndarray): Lion's current position
        current_f (float): Lion's current fitness
        previous_iter_best_f (float): Lion's fitness at end of previous iteration.
        has_improved (bool): Lion has improved fitness since last iteration.

    See Also:
        * :class:`niapy.algorithms.Individual`

    """

    def __init__(self, gender="m", has_pride=False, pride=-1, hunting_group=0, has_improved=True, **kwargs):
        r"""Initialize the Lion.

        Args:
            gender (Optional[string]): Lion's gender.
            has_pride (Optional[bool]): Lion has a pride.
            pride (Optional[int]): Lion's pride id.
            hunting_group (Optional[int]): Lion's hunting group id.
            has_improved (Optional[bool]): Lion has improved fitness since last iteration.

        See Also:
            * :func:`niapy.algorithms.Individual.__init__`

        """
        super().__init__(**kwargs)
        self.gender = gender
        self.has_pride = has_pride
        self.pride = pride
        self.hunting_group = hunting_group
        self.current_x = np.copy(self.x)
        self.current_f = self.f
        self.previous_iter_best_f = self.f
        self.has_improved = has_improved


class LionOptimizationAlgorithm(Algorithm):
    r"""Implementation of lion optimization algorithm.

    Algorithm:
        Lion Optimization algorithm

    Date:
        2021

    Authors:
        Aljoša Mesarec

    License:
        MIT

    Reference URL:
        https://doi.org/10.1016/j.jcde.2015.06.003

    Reference paper:
        Yazdani, Maziar, Jolai, Fariborz. Lion Optimization Algorithm (LOA): A nature-inspired metaheuristic algorithm. Journal of Computational Design and Engineering, Volume 3, Issue 1, Pages 24-36. 2016.

    Attributes:
        Name (List[str]): List of strings representing name of the algorithm.
        population_size (Optional[int]): Population size :math:`\in [1, \infty)`.
        nomad_ratio (Optional[float]): Ratio of nomad lions :math:`\in [0, 1]`.
        num_of_prides = Number of prides :math:`\in [1, \infty)`.
        female_ratio = Ratio of female lions in prides :math:`\in [0, 1]`.
        roaming_factor = Roaming factor :math:`\in [0, 1]`.
        mating_factor = Mating factor :math:`\in [0, 1]`.
        mutation_factor = Mutation factor :math:`\in [0, 1]`.
        immigration_factor = Immigration factor :math:`\in [0, 1]`.

    See Also:
        * :class:`niapy.algorithms.Algorithm`

    """

    Name = ['LionOptimizationAlgorithm', 'LOA']

    @staticmethod
    def info():
        r"""Get information about algorithm.

        Returns:
            str: Algorithm information

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r'''Yazdani, Maziar, Jolai, Fariborz. Lion Optimization Algorithm (LOA): A nature-inspired metaheuristic algorithm. Journal of Computational Design and Engineering, Volume 3, Issue 1, Pages 24-36. 2016.'''

    def __init__(self, population_size=50, nomad_ratio=0.2, num_of_prides=5, female_ratio=0.8, roaming_factor=0.2, mating_factor=0.3, mutation_factor=0.2, immigration_factor=0.4, *args, **kwargs):
        r"""Initialize LionOptimizationAlgorithm.

        Args:
            population_size (Optional[int]): Population size :math:`\in [1, \infty)`.
            nomad_ratio (Optional[float]): Ratio of nomad lions :math:`\in [0, 1]`.
            num_of_prides = Number of prides :math:`\in [1, \infty)`.
            female_ratio = Ratio of female lions in prides :math:`\in [0, 1]`.
            roaming_factor = Roaming factor :math:`\in [0, 1]`.
            mating_factor = Mating factor :math:`\in [0, 1]`.
            mutation_factor = Mutation factor :math:`\in [0, 1]`.
            immigration_factor = Immigration factor :math:`\in [0, 1]`.


        See Also:
            * :func:`niapy.algorithms.Algorithm.__init__`

        """
        super().__init__(population_size, individual_type=kwargs.pop('individual_type', Lion),
                         initialization_function=kwargs.pop('initialization_function', default_individual_init), *args, **kwargs)
        self.nomad_ratio = nomad_ratio
        self.num_of_prides = num_of_prides
        self.female_ratio = female_ratio
        self.roaming_factor = roaming_factor
        self.mating_factor = mating_factor
        self.mutation_factor = mutation_factor
        self.immigration_factor = immigration_factor

    def set_parameters(self, population_size=50, nomad_ratio=0.2, num_of_prides=5, female_ratio=0.8, roaming_factor=0.2, mating_factor=0.3, mutation_factor=0.2, immigration_factor=0.4, **kwargs):
        r"""Set the arguments of an algorithm.

        Args:
            population_size (Optional[int]): Population size :math:`\in [1, \infty)`.
            nomad_ratio (Optional[float]): Ratio of nomad lions :math:`\in [0, 1]`.
            num_of_prides = Number of prides :math:`\in [1, \infty)`.
            female_ratio = Ratio of female lions in prides :math:`\in [0, 1]`.
            roaming_factor = Roaming factor :math:`\in [0, 1]`.
            mating_factor = Mating factor :math:`\in [0, 1]`.
            mutation_factor = Mutation factor :math:`\in [0, 1]`.
            immigration_factor = Immigration factor :math:`\in [0, 1]`.

        See Also:
            * :func:`niapy.algorithms.Algorithm.set_parameters`

        """
        super().set_parameters(population_size=population_size, individual_type=Lion,
                               initialization_function=kwargs.pop('initialization_function', default_individual_init()), **kwargs)
        self.nomad_ratio = nomad_ratio
        self.num_of_prides = num_of_prides
        self.female_ratio = female_ratio
        self.roaming_factor = roaming_factor
        self.mating_factor = mating_factor
        self.mutation_factor = mutation_factor
        self.immigration_factor = immigration_factor

    def get_parameters(self):
        r"""Get parameters of the algorithm.

        Returns:
            Dict[str, Any]: Algorithm Parameters.

        """
        d = super().get_parameters()
        d.update({
            'nomad_ratio': self.nomad_ratio,
            'num_of_prides': self.num_of_prides,
            'female_ratio': self.female_ratio,
            'roaming_factor': self.roaming_factor,
            'mating_factor': self.mating_factor,
            'mutation_factor': self.mutation_factor,
            'immigration_factor': self.immigration_factor
        })
        return d

    def init_population(self, task):
        r"""Initialize starting population.

        Args:
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray[Lion], numpy.ndarray[float], Dict[str, Any]]:
                1. Initialized population of lions.
                2. Initialized populations function/fitness values.
                3. Additional arguments:
                    * pride_size (numpy.ndarray): Pride and nomad sizes.
                    * gender_distribution (numpy.ndarray): Pride and nomad gender distributions.

        """
        pop, fpop, d = super().init_population(task)
        pop, d = self.init_population_data(pop, d)
        return pop, fpop, d

    def init_population_data(self, pop, d):
        r"""Initialize data of starting population.

        Args:
            pop (numpy.ndarray[Lion]: Starting lion population
            d (Dict[str, Any]): Additional arguments

        Returns:
            Tuple[numpy.ndarray[Lion], Dict[str, Any]]:
                1. Initialized population of lions.
                2. Additional arguments:
                    * pride_size (numpy.ndarray): Pride and nomad sizes.
                    * gender_distribution (numpy.ndarray): Pride and nomad gender distributions.

        """
        nomad_size = round(self.nomad_ratio * self.population_size)

        # Creating array of pride sizes.
        pride_size = np.zeros(self.num_of_prides + 1, dtype=int)
        pride_size[-1] = nomad_size
        remaining_lions = self.population_size - nomad_size
        while(remaining_lions > 0):
            if(remaining_lions >= self.num_of_prides):
                pride_size[:self.num_of_prides] += 1
            else:
                pride_size[:remaining_lions] += 1
            remaining_lions -= self.num_of_prides

        # Setting gender and pride id of pride members.
        index_counter = 0
        for i in range(self.num_of_prides):
            curr_pride_size = pride_size[i]
            num_of_females = round(self.female_ratio * curr_pride_size)
            for lion in pop[index_counter:index_counter + curr_pride_size]:
                lion.has_pride = True
                lion.pride = i
                if(num_of_females > 0):
                    lion.gender = "f"
                    num_of_females -= 1
            index_counter += curr_pride_size

        # Setting gender of nomads
        num_of_females = round((1 - self.female_ratio) * pride_size[-1])
        for lion in pop[index_counter:index_counter + num_of_females]:
            lion.gender = "f"

        # Creating array of pride gender quantities
        gender_distribution = np.zeros((self.num_of_prides + 1, 2), dtype=int)
        index_counter = 0
        for i in range(self.num_of_prides):
            curr_pride_size = pride_size[i]
            for lion in pop[index_counter:index_counter + curr_pride_size]:
                if lion.gender == "f":
                    gender_distribution[i][0] += 1
                elif lion.gender == "m":
                    gender_distribution[i][1] += 1
            index_counter += curr_pride_size
        # Creating array of nomad gender quantities
        for lion in pop[index_counter:]:
            if lion.gender == "f":
                gender_distribution[self.num_of_prides][0] += 1
            elif lion.gender == "m":
                gender_distribution[self.num_of_prides][1] += 1

        # Updating params
        d.update({'pride_size': pride_size, 'gender_distribution': gender_distribution})
        return pop, d

    def hunting(self, population, pride_size, task):
        r"""Pride female hunters go hunting.

        Args:
            population (numpy.ndarray[Lion]): Lion population.
            pride_size (numpy.ndarray[int]): Pride and nomad sizes.
            task (Task): Optimization task.

        Returns:
            population (numpy.ndarray[Lion]): Lion population that finished with hunting.

        """
        num_of_prides = len(pride_size) - 1
        index_counter_pride = 0
        for pride_i in range(num_of_prides):
            curr_pride_size = pride_size[pride_i]
            prey_x = np.zeros(task.dimension, dtype=float)
            num_of_hunters = 0
            hunting_group_fitness = np.zeros(4)
            for lion in population[index_counter_pride:index_counter_pride + curr_pride_size]:
                if lion.gender == "f":
                    lion.hunting_group = self.integers(0, 4)
                    hunting_group_fitness[lion.hunting_group] += lion.current_f
                if not (lion.hunting_group == 0):
                    prey_x += lion.current_x
                    num_of_hunters += 1

            # Group with highest fitness becomes center group, the rest become left and right groups
            sorted_hunting_group_indices = np.argsort(hunting_group_fitness[1:])
            right_group, left_group, center_group = sorted_hunting_group_indices + 1

            # Prey's position is average position of hunters.
            if not (num_of_hunters == 0):
                prey_x /= num_of_hunters

            # Check if prey's new position is in limits.
            prey_x = task.repair(prey_x)

            # Calculate new positions of hunters with "Opposition-Based Learning method".
            for lion in population[index_counter_pride:index_counter_pride + curr_pride_size]:
                if lion.hunting_group == left_group or lion.hunting_group == right_group:
                    for i in range(task.dimension):
                        if (2 * prey_x[i] - lion.current_x[i]) < prey_x[i]:
                            lion.current_x[i] = self.uniform((2 * prey_x[i] - lion.current_x[i]), prey_x[i])
                        elif (2 * prey_x[i] - lion.current_x[i]) > prey_x[i]:
                            lion.current_x[i] = self.uniform(prey_x[i], (2 * prey_x[i] - lion.current_x[i]))
                if lion.hunting_group == center_group:
                    for i in range(task.dimension):
                        if lion.current_x[i] < prey_x[i]:
                            lion.current_x[i] = self.uniform(lion.current_x[i], prey_x[i])
                        elif lion.current_x[i] > prey_x[i]:
                            lion.current_x[i] = self.uniform(prey_x[i], lion.current_x[i])

                if not(lion.hunting_group == 0):
                    # Check if lion's new position is in limits.
                    lion.current_x = task.repair(lion.current_x)
                    lion.current_f = task.eval(lion.current_x)
                    # If hunter's new fitness is better then change prey's position.
                    if lion.current_f < lion.f:
                        lion.x = np.copy(lion.current_x)
                        lion.f = lion.current_f
                        percentage_of_improvement = 1 - lion.f / lion.previous_iter_best_f
                        prey_x = prey_x + self.random() * percentage_of_improvement * (prey_x - lion.current_x)
                        # Check if prey's new position is in limits.
                        prey_x = task.repair(prey_x)

            index_counter_pride += curr_pride_size

        return population

    def move_to_safe_place(self, population, pride_size, task):
        r"""Female pride lions move towards position with good fitness.

        Args:
            population (numpy.ndarray[Lion]): Lion population.
            pride_size (numpy.ndarray[int]): Pride and nomad sizes.
            task (Task): Optimization task.

        Returns:
            population (numpy.ndarray[Lion]): Lion population that finished with moving to safe place.

        """
        num_of_prides = len(pride_size) - 1
        index_counter_pride = 0
        for pride_i in range(num_of_prides):
            curr_pride_size = pride_size[pride_i]
            num_of_improvements = 0
            pride_territory = []
            for lion in population[index_counter_pride:index_counter_pride + curr_pride_size]:
                lion_copy = copy.deepcopy(lion)
                pride_territory = np.append(pride_territory, objects_to_array([lion_copy]))
                if lion.has_improved:
                    num_of_improvements += 1
            # Tournament selection to select places in territory if there's more than 2 places
            if len(pride_territory) > 1:
                tournament_size = max(2, int(np.ceil(num_of_improvements / 2)))
                tournament_selections = self.rng.choice(pride_territory, tournament_size, replace=False)
                tournament_winner = tournament_selections[0].x.copy()
                tournament_min_f = tournament_selections[0].f
                for candidate in tournament_selections[1:]:
                    if candidate.f < tournament_min_f:
                        tournament_min_f = candidate.f
                        tournament_winner = candidate.x.copy()
            else:
                tournament_winner = pride_territory[0].x.copy()
                tournament_min_f = pride_territory[0].f

            # Move female non-hunters
            for lion in population[index_counter_pride:index_counter_pride + curr_pride_size]:
                if lion.gender == "f" and lion.hunting_group == 0:
                    # Get vector r_one.
                    r_one = tournament_winner.copy()
                    r_one -= lion.x
                    # Get vector 2_two with Gram-Schmidt process.
                    if np.linalg.norm((r_one).T) == 0:
                        # If r_one vector is 0 then Gram-Schmidt process return wrong values
                        r_two = np.zeros(len(r_one.T))
                        rand_index = self.integers(0, len(r_one.T))
                        r_two[rand_index] = 1
                    else:
                        # Gram-Schmidt process to find orthogonal vector r_two.
                        random_vec = self.standard_normal(len(r_one))
                        r_two = random_vec - ((r_one.T).dot(random_vec)) / ((r_one.T).dot(r_one)) * r_one
                    # Calculate other variables and new lion's position
                    d = np.linalg.norm(r_one) / np.linalg.norm(task.upper[0] - task.lower[0])
                    rnd_num = self.random()
                    rnd_num_u = self.uniform(-1, 1)
                    angle = self.uniform(-np.pi / 6, np.pi / 6)
                    lion.current_x += 2 * d * rnd_num * r_one + rnd_num_u * np.tan(angle) * d * r_two
                    # Check if lion's current position is in limits.
                    lion.current_x = task.repair(lion.current_x)
                    lion.current_f = task.eval(lion.current_x)
                    # If lion's position has improved update best position and fitness
                    if lion.current_f < lion.f:
                        lion.x = np.copy(lion.current_x)
                        lion.f = lion.current_f
            index_counter_pride += curr_pride_size

        return population

    def roaming(self, population, pride_size, task):
        r"""Male lions move towards new position.

        Args:
            population (numpy.ndarray[Lion]): Lion population.
            pride_size (numpy.ndarray[int]): Pride and nomad sizes.
            task (Task): Optimization task.

        Returns:
            population (numpy.ndarray[Lion]): Lion population that finished with roaming.

        """
        num_of_prides = len(pride_size) - 1
        index_counter_pride = 0
        # Pride lions roam.
        for pride_i in range(num_of_prides):
            curr_pride_size = pride_size[pride_i]
            for lion in population[index_counter_pride:index_counter_pride + curr_pride_size]:
                if lion.gender == "m":
                    # Select all lions in pride.
                    pride_lions = []
                    for p_l in population[index_counter_pride:index_counter_pride + curr_pride_size]:
                        lion_copy = copy.deepcopy(p_l)
                        pride_lions = np.append(pride_lions, objects_to_array([lion_copy]))
                    # Select random lions, their amount is based on roaming factor.
                    num_of_selected_lions = round(len(pride_lions) * self.roaming_factor)
                    selected_lions = self.rng.choice(pride_lions, num_of_selected_lions, replace=False)
                    # Move towards territories of selected lions
                    for selected_lion in selected_lions:
                        d = np.linalg.norm(selected_lion.x - lion.x) / np.linalg.norm(task.upper[0] - task.lower[0])
                        x = self.uniform(0, 2 * d)
                        angle = self.uniform(-np.pi / 6, np.pi / 6)
                        lion.current_x += x * d * np.tan(angle)
                        # Check if lion's new position is in limits.
                        lion.current_x = task.repair(lion.current_x)
                        lion.current_f = task.eval(lion.current_x)
                        # Update best position/fitness if lion's best position is improved
                        if lion.current_f < lion.f:
                            lion.x = np.copy(lion.current_x)
                            lion.f = lion.current_f

            index_counter_pride += curr_pride_size

        # Nomad lions roam.
        nomad_size = pride_size[-1]
        for lion in population[len(population) - nomad_size:]:
            best_nomad_fitness = np.min([c_l.current_f for c_l in population[len(population) - nomad_size:]])
            roaming_probability = 0.1 + np.minimum(0.5, (lion.current_f - best_nomad_fitness) / best_nomad_fitness)
            # If roaming threshold is met, move lion to a random new position.
            if self.random() <= roaming_probability:
                lion.current_x = self.uniform(task.lower, task.upper, task.dimension)
                lion.current_f = task.eval(lion.current_x)
            # Update best position/fitness if lion's best position is improved
            if lion.current_f < lion.f:
                lion.x = np.copy(lion.current_x)
                lion.f = lion.current_f

        return population

    def mating(self, population, pride_size, gender_distribution, task):
        r"""Female lions mate with male lions to produce offspring.

        Args:
            population (numpy.ndarray[Lion]): Lion population.
            pride_size (numpy.ndarray[int]): Pride and nomad sizes.
            gender_distribution (numpy.ndarray[int]): Pride and nomad gender distribution.
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray[Lion], numpy.ndarray[int]):
                1. Lion population that finished with mating.
                2. Pride and nomad excess gender quantities.

        """
        added_cubs = []
        excess_lion_gender_quantities = np.zeros((self.num_of_prides + 1, 2), dtype=int)
        num_of_prides = len(pride_size) - 1
        # Copy of all pride lions.
        pride_lions = []
        index_counter_pride = 0
        for pride_i in range(num_of_prides):
            curr_pride_size = pride_size[pride_i]
            for lion in population[index_counter_pride:index_counter_pride + curr_pride_size]:
                lion_copy = copy.deepcopy(lion)
                pride_lions = np.append(pride_lions, objects_to_array([lion_copy]))
            index_counter_pride += curr_pride_size

        # Copy of all nomad lions.
        nomad_lions = []
        nomad_size = pride_size[-1]
        for lion in population[len(population) - nomad_size]:
            lion_copy = copy.deepcopy(lion)
            nomad_lions = np.append(nomad_lions, objects_to_array([lion_copy]))

        # Prides mating.
        index_counter_pride = 0
        for pride_i in range(num_of_prides):
            curr_pride_size = pride_size[pride_i]
            num_of_males = gender_distribution[pride_i][1]
            # If there's at least 1 male, proceed
            if not num_of_males == 0:
                # Array of males.
                males = []
                for lion in population[index_counter_pride:index_counter_pride + curr_pride_size]:
                    if lion.gender == "m":
                        lion_copy = copy.deepcopy(lion)
                        males = np.append(males, objects_to_array([lion_copy]))
                # Mate all females with a mating probability.
                for lion in population[index_counter_pride:index_counter_pride + curr_pride_size]:
                    if lion.gender == "f" and self.random() < self.mating_factor:
                        # Choose males that will mate.
                        num_of_mating_males = self.integers(1, num_of_males)
                        mating_males = self.rng.choice(males, num_of_mating_males, replace=False)

                        beta = self.normal(0.5, 0.1)

                        # Total position x of mating males. Needed for calculating average.
                        mating_males_x_sum = np.zeros(task.dimension)
                        for mating_male in mating_males:
                            mating_males_x_sum = np.add(mating_males_x_sum, mating_male.x)
                        # Calculate position x for offsprings.
                        offspring_one_position = beta * lion.x + ((1 - beta) * mating_males_x_sum / num_of_mating_males)
                        offspring_two_position = (1 - beta) * lion.x + (beta * mating_males_x_sum / num_of_mating_males)

                        # Mutation of the genes with mutation probability.
                        for i in range(task.dimension):
                            if self.random() < self.mutation_factor:
                                offspring_one_position[i] = self.uniform(task.lower[i], task.upper[i], 1)
                            if self.random() < self.mutation_factor:
                                offspring_two_position[i] = self.uniform(task.lower[i], task.upper[i], 1)
                        # Create offspring Lion objects
                        offspring_one = copy.deepcopy(lion)
                        offspring_two = copy.deepcopy(lion)
                        offspring_one.has_pride = True
                        offspring_two.has_pride = True
                        offspring_one.pride = pride_i
                        offspring_two.pride = pride_i
                        offspring_one.hunting_group = 0
                        offspring_two.hunting_group = 0
                        offspring_one.has_improved = True
                        offspring_two.has_improved = True
                        # Randomly assign genders to offsprings.
                        if self.random() < 0.5:
                            offspring_one.gender = "m"
                            offspring_two.gender = "f"
                        else:
                            offspring_one.gender = "f"
                            offspring_two.gender = "m"
                        offspring_one.x = offspring_one_position
                        offspring_two.x = offspring_two_position
                        # Check if offspring's position is in limits.
                        offspring_one.evaluate(task)
                        offspring_two.evaluate(task)
                        # Assign other offspring's values.
                        offspring_one.current_x = np.copy(offspring_one.x)
                        offspring_two.current_x = np.copy(offspring_two.x)
                        offspring_one.current_f = offspring_one.f
                        offspring_two.current_f = offspring_two.f
                        offspring_one.previous_iter_best_f = offspring_one.f + 1
                        offspring_two.previous_iter_best_f = offspring_two.f + 1
                        # Add offspring to array of added cubs.
                        added_cubs = np.append(added_cubs, objects_to_array([offspring_one]))
                        added_cubs = np.append(added_cubs, objects_to_array([offspring_two]))
                        excess_lion_gender_quantities[pride_i][0] += 1
                        excess_lion_gender_quantities[pride_i][1] += 1
                        pride_size[pride_i] += 2
                        gender_distribution[pride_i][0] += 1
                        gender_distribution[pride_i][1] += 1
            index_counter_pride += curr_pride_size

        # Nomads mating.
        nomad_size = pride_size[-1]
        num_of_males = gender_distribution[pride_i][1]
        # If there's at least one male nomad, proceed
        if not num_of_males == 0:
            # Create array of males that can mate.
            males = []
            for lion in population[len(population) - nomad_size:]:
                if lion.gender == "m":
                    lion_copy = copy.deepcopy(lion)
                    males = np.append(males, objects_to_array([lion_copy]))
            # Mate all females with mating probability.
            for lion in population[len(population) - nomad_size:]:
                if lion.gender == "f" and self.random() < self.mating_factor:
                    # Choose one male that will mate.
                    mating_male = self.rng.choice(males)

                    beta = self.normal(0.5, 0.1)

                    # Calculate x for offsprings.
                    offspring_one_position = beta * lion.x + ((1 - beta) * mating_male.x)
                    offspring_two_position = (1 - beta) * lion.x + beta * mating_male.x

                    # Mutation of the genes with mutation probability.
                    for i in range(task.dimension):
                        if self.random() < self.mutation_factor:
                            offspring_one_position[i] = self.uniform(task.lower[i], task.upper[i])
                        if self.random() < self.mutation_factor:
                            offspring_two_position[i] = self.uniform(task.lower[i], task.upper[i])

                    # Create offspring Lion objects.
                    offspring_one = copy.deepcopy(lion)
                    offspring_two = copy.deepcopy(lion)
                    offspring_one.has_pride = False
                    offspring_two.has_pride = False
                    offspring_one.pride = -1
                    offspring_two.pride = -1
                    offspring_one.hunting_group = 0
                    offspring_two.hunting_group = 0
                    offspring_one.has_improved = True
                    offspring_two.has_improved = True
                    # Randomly assign genders to offsprings.
                    if self.random() < 0.5:
                        offspring_one.gender = "m"
                        offspring_two.gender = "f"
                    else:
                        offspring_one.gender = "f"
                        offspring_two.gender = "m"

                    offspring_one.x = offspring_one_position
                    offspring_two.x = offspring_two_position
                    # Check if offspring's position is in limits.
                    offspring_one.evaluate(task)
                    offspring_two.evaluate(task)
                    # Assign other offspring's values.
                    offspring_one.current_x = np.copy(offspring_one.x)
                    offspring_two.current_x = np.copy(offspring_two.x)
                    offspring_one.current_f = offspring_one.f
                    offspring_two.current_f = offspring_two.f
                    offspring_one.previous_iter_best_f = offspring_one.f + 1
                    offspring_two.previous_iter_best_f = offspring_two.f + 1
                    # Add offspring to array of added cubs
                    added_cubs = np.append(added_cubs, objects_to_array([offspring_one]))
                    added_cubs = np.append(added_cubs, objects_to_array([offspring_two]))
                    excess_lion_gender_quantities[-1][0] += 1
                    excess_lion_gender_quantities[-1][1] += 1
                    pride_size[-1] += 2
                    gender_distribution[-1][0] += 1
                    gender_distribution[-1][1] += 1

        # Add pride originals and cubs to same population.
        new_population = []
        original_index_counter_pride = 0
        cub_index_counter_pride = 0
        for pride_i in range(num_of_prides):
            # Append original pride lion.
            curr_original_pride_size = pride_size[pride_i] - excess_lion_gender_quantities[pride_i][0] - excess_lion_gender_quantities[pride_i][1]
            for lion in population[original_index_counter_pride:original_index_counter_pride + curr_original_pride_size]:
                lion_copy = copy.deepcopy(lion)
                new_population = np.append(new_population, objects_to_array([lion_copy]))
            # Append cub pride lions.
            curr_cub_pride_size = excess_lion_gender_quantities[pride_i][0] + excess_lion_gender_quantities[pride_i][1]
            for lion in added_cubs[cub_index_counter_pride:cub_index_counter_pride + curr_cub_pride_size]:
                lion_copy = copy.deepcopy(lion)
                new_population = np.append(new_population, objects_to_array([lion_copy]))

            original_index_counter_pride += curr_original_pride_size
            cub_index_counter_pride += curr_cub_pride_size
        # Add nomad originals and cubs to same population.
        originals_nomad_size = pride_size[-1] - excess_lion_gender_quantities[-1][0] - excess_lion_gender_quantities[-1][1]
        for lion in population[len(population) - originals_nomad_size:]:
            lion_copy = copy.deepcopy(lion)
            new_population = np.append(new_population, objects_to_array([lion_copy]))
        cubs_nomad_size = excess_lion_gender_quantities[-1][0] + excess_lion_gender_quantities[-1][1]
        for lion in added_cubs[len(added_cubs) - cubs_nomad_size:]:
            lion_copy = copy.deepcopy(lion)
            new_population = np.append(new_population, objects_to_array([lion_copy]))

        return new_population, excess_lion_gender_quantities

    def defense(self, population, pride_size, gender_distribution, excess_lion_gender_quantities, task):
        r"""Male lions attack other lions in pride.

        Args:
            population (numpy.ndarray[Lion]): Lion population.
            pride_size (numpy.ndarray[int]): Pride and nomad sizes.
            gender_distribution (numpy.ndarray[int]): Pride and nomad gender distribution.
            excess_lion_gender_quantities (numpy.ndarray[int]): Pride and nomad excess members.
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray[Lion], numpy.ndarray[int]):
                1. Lion population that finished with defending.
                2. Pride and nomad excess gender quantities.

        """
        new_nomads = []
        original_pride_lions = []
        num_of_prides = len(pride_size) - 1
        # Pride lions defense
        index_counter_pride = 0
        for pride_i in range(num_of_prides):
            curr_pride_size = pride_size[pride_i]
            num_of_males_to_be_kicked = excess_lion_gender_quantities[pride_i][1]
            males = []
            # Go through pride
            for lion in population[index_counter_pride:index_counter_pride + curr_pride_size]:
                lion_copy = copy.deepcopy(lion)
                if lion.gender == "m":
                    males = np.append(males, objects_to_array([lion_copy]))
                elif lion.gender == "f":
                    original_pride_lions = np.append(original_pride_lions, objects_to_array([lion_copy]))
            # Find males with worst fitness that will be kicked, leave the rest
            males = sorted(males, key=lambda lion: lion.current_f, reverse=True)
            for lion in males:
                lion_copy = copy.deepcopy(lion)
                if num_of_males_to_be_kicked == 0:
                    original_pride_lions = np.append(original_pride_lions, objects_to_array([lion_copy]))
                else:
                    new_nomads = np.append(new_nomads, objects_to_array([lion_copy]))
                    num_of_males_to_be_kicked -= 1
            index_counter_pride += curr_pride_size

        # Create new population after kicking pride lions
        moved_population = []
        # Append original pride lions.
        for lion in original_pride_lions:
            lion_copy = copy.deepcopy(lion)
            moved_population = np.append(moved_population, objects_to_array([lion_copy]))
        # Append original nomads.
        original_nomads_size = pride_size[-1]
        for lion in population[len(population) - original_nomads_size:]:
            lion_copy = copy.deepcopy(lion)
            moved_population = np.append(moved_population, objects_to_array([lion_copy]))
        # Append new nomads.
        for lion in new_nomads:
            lion_copy = copy.deepcopy(lion)

            excess_lion_gender_quantities[lion_copy.pride][1] -= 1
            gender_distribution[lion_copy.pride][1] -= 1
            pride_size[lion_copy.pride] -= 1

            lion_copy.has_pride = False
            lion_copy.pride = -1

            moved_population = np.append(moved_population, objects_to_array([lion_copy]))

            excess_lion_gender_quantities[-1][1] += 1
            gender_distribution[-1][1] += 1
            pride_size[-1] += 1

        # Nomad lions defense.
        nomads_size = pride_size[-1]
        for nomad_lion in moved_population[len(moved_population) - nomads_size:]:
            nomad_lion_has_won = False
            # Create binary template - which prides a nomad lion will attack.
            pride_index_to_attack = np.zeros(num_of_prides, dtype=int)
            for i in range(num_of_prides):
                if self.random() < 0.5:
                    pride_index_to_attack[i] = 1
            # Nomad attacks prides based on binary template.
            index_counter_pride = 0
            for pride_i in range(num_of_prides):
                curr_pride_size = pride_size[pride_i]
                if pride_index_to_attack[pride_i] == 1:
                    # Attack all male lions.
                    for pride_lion in moved_population[index_counter_pride:index_counter_pride + curr_pride_size]:
                        if(lion.gender == "m"):
                            # Swap nomad and pride lion if nomad has better fitness.
                            if nomad_lion.current_f < pride_lion.current_f:
                                copy_nomad_lion = copy.deepcopy(nomad_lion)
                                copy_pride_lion = copy.deepcopy(pride_lion)

                                pride_lion = copy_nomad_lion
                                pride_lion.has_pride = True
                                pride_lion.pride = pride_i

                                nomad_lion = copy_pride_lion
                                nomad_lion.has_pride = False
                                nomad_lion.pride = -1
                                # If nomad lion won the attack, we continue with next nomad.
                                nomad_lion_has_won = True
                                break
                    if nomad_lion_has_won:
                        break
                index_counter_pride += curr_pride_size

        return moved_population, excess_lion_gender_quantities

    def migration(self, population, pride_size, gender_distribution, excess_lion_gender_quantities, task):
        r"""Female lions randomly become nomad.

        Args:
            population (numpy.ndarray[Lion]): Lion population.
            pride_size (numpy.ndarray[int]): Pride and nomad sizes.
            gender_distribution (numpy.ndarray[int]): Pride and nomad gender distribution.
            excess_lion_gender_quantities (numpy.ndarray[int]): Pride and nomad excess members.
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray[Lion], numpy.ndarray[int]):
                1. Lion population that finished with migration.
                2. Pride and nomad excess gender quantities.

        """
        new_nomads = []
        original_pride_lions = []

        num_of_prides = len(pride_size) - 1
        # Pride females migration
        index_counter_pride = 0
        for pride_i in range(num_of_prides):
            curr_pride_size = pride_size[pride_i]
            num_of_females = gender_distribution[pride_i][0]
            num_of_excess_females = excess_lion_gender_quantities[pride_i][0]
            num_of_females_to_migrate = num_of_excess_females + round(((num_of_females - num_of_excess_females) * self.immigration_factor))

            females = []
            # Go through pride
            for lion in population[index_counter_pride:index_counter_pride + curr_pride_size]:
                lion_copy = copy.deepcopy(lion)
                if lion.gender == "m":
                    original_pride_lions = np.append(original_pride_lions, objects_to_array([lion_copy]))
                elif lion.gender == "f":
                    females = np.append(females, objects_to_array([lion_copy]))
            # Migrate random females, leave the rest
            females_indices_to_migrate = np.zeros(num_of_females, dtype=int)
            for i in range(num_of_females_to_migrate):
                females_indices_to_migrate[i] = 1
            self.rng.shuffle(females_indices_to_migrate)

            for i, lion in enumerate(females):
                lion_copy = copy.deepcopy(lion)
                if females_indices_to_migrate[i] == 1:
                    new_nomads = np.append(new_nomads, objects_to_array([lion_copy]))
                else:
                    original_pride_lions = np.append(original_pride_lions, objects_to_array([lion_copy]))
            index_counter_pride += curr_pride_size

        # Create new population after migrating pride lions
        moved_population = []
        # Append original pride lions
        for lion in original_pride_lions:
            lion_copy = copy.deepcopy(lion)
            moved_population = np.append(moved_population, objects_to_array([lion_copy]))
        # Append original nomads
        original_nomads_size = pride_size[-1]
        for lion in population[len(population) - original_nomads_size:]:
            lion_copy = copy.deepcopy(lion)
            moved_population = np.append(moved_population, objects_to_array([lion_copy]))
        # Append new nomads
        for lion in new_nomads:
            lion_copy = copy.deepcopy(lion)

            excess_lion_gender_quantities[lion_copy.pride][0] -= 1
            gender_distribution[lion_copy.pride][0] -= 1
            pride_size[lion_copy.pride] -= 1

            lion_copy.has_pride = False
            lion_copy.pride = -1

            moved_population = np.append(moved_population, objects_to_array([lion_copy]))

            excess_lion_gender_quantities[-1][0] += 1
            gender_distribution[-1][0] += 1
            pride_size[-1] += 1

        # Fill up empty female pride spaces with best nomad female lions
        prides_spots_to_be_filled = 0
        for i in range(num_of_prides):
            prides_spots_to_be_filled += np.abs(excess_lion_gender_quantities[i][0])

        nomad_females = []
        nomad_males = []
        original_nomads_size = pride_size[-1]
        for lion in moved_population[len(population) - original_nomads_size:]:
            lion_copy = copy.deepcopy(lion)
            if lion.gender == "f":
                nomad_females = np.append(nomad_females, objects_to_array([lion_copy]))
            elif lion.gender == "m":
                nomad_males = np.append(nomad_males, objects_to_array([lion_copy]))
        nomad_females = sorted(nomad_females, key=lambda lion: lion.current_f, reverse=False)
        nomad_females_to_move = []
        nomad_females_to_keep = []
        counter = prides_spots_to_be_filled
        for lion in nomad_females:
            lion_copy = copy.deepcopy(lion)
            if not counter == 0:
                nomad_females_to_move = np.append(nomad_females_to_move, objects_to_array([lion_copy]))
                counter -= 1
            else:
                nomad_females_to_keep = np.append(nomad_females_to_keep, objects_to_array([lion_copy]))
        self.rng.shuffle(nomad_females_to_move)

        # Append pride lions and moved female nomads
        final_population = []
        index_counter_pride = 0
        index_females_to_move = 0
        for pride_i in range(num_of_prides):
            curr_pride_size = pride_size[pride_i]
            # Append pride lions
            for lion in moved_population[index_counter_pride:index_counter_pride + curr_pride_size]:
                lion_copy = copy.deepcopy(lion)
                final_population = np.append(final_population, objects_to_array([lion_copy]))
            # Append female nomads
            curr_pride_spots_empty = np.abs(excess_lion_gender_quantities[pride_i][0])
            for lion in nomad_females_to_move[index_females_to_move:index_females_to_move + curr_pride_spots_empty]:
                lion_copy = copy.deepcopy(lion)

                excess_lion_gender_quantities[lion_copy.pride][0] -= 1
                gender_distribution[lion_copy.pride][0] -= 1
                pride_size[lion_copy.pride] -= 1

                lion_copy.has_pride = True
                lion_copy.pride = pride_i

                final_population = np.append(final_population, objects_to_array([lion_copy]))

                excess_lion_gender_quantities[pride_i][0] += 1
                gender_distribution[pride_i][0] += 1
                pride_size[pride_i] += 1
            # Increase starting indices for array search
            index_counter_pride += curr_pride_size
            index_females_to_move += curr_pride_spots_empty
        # Append the kept nomad females
        for lion in nomad_females_to_keep:
            lion_copy = copy.deepcopy(lion)
            final_population = np.append(final_population, objects_to_array([lion_copy]))
        # Append nomad males
        for lion in nomad_males:
            lion_copy = copy.deepcopy(lion)
            final_population = np.append(final_population, objects_to_array([lion_copy]))
        return final_population, excess_lion_gender_quantities

    def population_equilibrium(self, population, pride_size, gender_distribution, excess_lion_gender_quantities, task):
        r"""Remove extra nomad lions.

        Args:
            population (numpy.ndarray[Lion]): Lion population.
            pride_size (numpy.ndarray[int]): Pride and nomad sizes.
            gender_distribution (numpy.ndarray[int]): Pride and nomad gender distribution.
            excess_lion_gender_quantities (numpy.ndarray[int]): Pride and nomad excess members.
            task (Task): Optimization task.

        Returns:
            final_population (numpy.ndarray[Lion]): Lion population with removed extra nomads.

        """
        nomad_females = []
        nomad_males = []
        kept_nomads = []
        # Number of males and females that need to be removed.
        num_of_female_nomads_to_remove = excess_lion_gender_quantities[-1][0]
        num_of_male_nomads_to_remove = excess_lion_gender_quantities[-1][1]
        original_nomads_size = pride_size[-1]
        # Get lists of nomad males and females.
        for lion in population[len(population) - original_nomads_size:]:
            lion_copy = copy.deepcopy(lion)
            if lion.gender == "f":
                nomad_females = np.append(nomad_females, objects_to_array([lion_copy]))
            elif lion.gender == "m":
                nomad_males = np.append(nomad_males, objects_to_array([lion_copy]))
        # Sort lists descendingly.
        nomad_males = sorted(nomad_males, key=lambda lion: lion.current_f, reverse=True)
        nomad_females = sorted(nomad_females, key=lambda lion: lion.current_f, reverse=True)
        # Remove extra lions that have bad fitness, keep the rest.
        for lion in nomad_males[num_of_male_nomads_to_remove:]:
            lion_copy = copy.deepcopy(lion)
            kept_nomads = np.append(kept_nomads, objects_to_array([lion_copy]))
        for lion in nomad_females[num_of_female_nomads_to_remove:]:
            lion_copy = copy.deepcopy(lion)
            kept_nomads = np.append(kept_nomads, objects_to_array([lion_copy]))

        # Append pride lions to final population.
        final_population = []
        for lion in population[:len(population) - original_nomads_size]:
            lion_copy = copy.deepcopy(lion)
            final_population = np.append(final_population, objects_to_array([lion_copy]))
        # Append kept nomads to final population.
        for lion in kept_nomads:
            lion_copy = copy.deepcopy(lion)
            final_population = np.append(final_population, objects_to_array([lion_copy]))
        pride_size[-1] -= (num_of_female_nomads_to_remove + num_of_male_nomads_to_remove)
        gender_distribution[-1][0] -= num_of_female_nomads_to_remove
        gender_distribution[-1][1] -= num_of_male_nomads_to_remove

        return final_population

    def data_correction(self, population, pride_size, task):
        r"""Update lion's data if his position has improved since last iteration.

        Args:
            population (numpy.ndarray[Lion]): Lion population.
            pride_size (numpy.ndarray[int]): Pride and nomad sizes.
            task (Task): Optimization task.

        Returns:
            population (numpy.ndarray[Lion]): Lion population with corrected data.

        """
        for lion in population:
            if lion.f < lion.previous_iter_best_f:
                lion.has_improved = True
                lion.previous_iter_best_f = lion.f
            else:
                lion.has_improved = False
        return population

    def run_iteration(self, task, population, population_fitness, best_x, best_fitness, **params):
        pride_size = params.pop('pride_size')
        gender_distribution = params.pop('gender_distribution')
        # Algorithm steps
        lions = self.hunting(population, pride_size, task)
        lions = self.move_to_safe_place(lions, pride_size, task)
        lions = self.roaming(lions, pride_size, task)
        lions, excess_lion_gender_quantities = self.mating(lions, pride_size, gender_distribution, task)
        lions, excess_lion_gender_quantities = self.defense(lions, pride_size, gender_distribution, excess_lion_gender_quantities, task)
        lions, excess_lion_gender_quantities = self.migration(lions, pride_size, gender_distribution, excess_lion_gender_quantities, task)
        lions = self.population_equilibrium(lions, pride_size, gender_distribution, excess_lion_gender_quantities, task)
        lions = self.data_correction(lions, pride_size, task)

        lions_fitness = np.asarray([lion.f for lion in lions])
        best_x, best_fitness = self.get_best(lions, lions_fitness, best_x, best_fitness)
        return lions, lions_fitness, best_x, best_fitness, {'pride_size': pride_size, 'gender_distribution': gender_distribution}



# encoding=utf8
import logging

import numpy as np

from niapy.algorithms.algorithm import Algorithm
from niapy.util.distances import euclidean

__all__ = ['GravitationalSearchAlgorithm']

logging.basicConfig()
logger = logging.getLogger('niapy.algorithms.basic')
logger.setLevel('INFO')


class GravitationalSearchAlgorithm(Algorithm):
    r"""Implementation of Gravitational Search Algorithm.

    Algorithm:
        Gravitational Search Algorithm

    Date:
        2018

    Author:
        Klemen Berkovič

    License:
        MIT

    Reference URL:
        https://doi.org/10.1016/j.ins.2009.03.004

    Reference paper:
        Esmat Rashedi, Hossein Nezamabadi-pour, Saeid Saryazdi, GSA: A Gravitational Search Algorithm, Information Sciences, Volume 179, Issue 13, 2009, Pages 2232-2248, ISSN 0020-0255

    Attributes:
        Name (List[str]): List of strings representing algorithm name.

    See Also:
        * :class:`niapy.algorithms.algorithm.Algorithm`

    """

    Name = ['GravitationalSearchAlgorithm', 'GSA']

    @staticmethod
    def info():
        r"""Get algorithm information.

        Returns:
            str: Algorithm information.

        """
        return r"""Esmat Rashedi, Hossein Nezamabadi-pour, Saeid Saryazdi, GSA: A Gravitational Search Algorithm, Information Sciences, Volume 179, Issue 13, 2009, Pages 2232-2248, ISSN 0020-0255"""

    def __init__(self, population_size=40, g0=2.467, epsilon=1e-17, *args, **kwargs):
        """Initialize GravitationalSearchAlgorithm.

        Args:
            population_size (Optional[int]): Population size.
            g0 (Optional[float]): Starting gravitational constant.
            epsilon (Optional[float]): Small number.

        See Also:
            * :func:`niapy.algorithms.algorithm.Algorithm.__init__`

        """
        super().__init__(population_size, *args, **kwargs)
        self.g0 = g0
        self.epsilon = epsilon

    def set_parameters(self, population_size=40, g0=2.467, epsilon=1e-17, **kwargs):
        r"""Set the algorithm parameters.

        Args:
            population_size (Optional[int]): Population size.
            g0 (Optional[float]): Starting gravitational constant.
            epsilon (Optional[float]): Small number.

        See Also:
            * :func:`niapy.algorithms.algorithm.Algorithm.set_parameters`

        """
        super().set_parameters(population_size=population_size, **kwargs)
        self.g0 = g0
        self.epsilon = epsilon

    def get_parameters(self):
        r"""Get algorithm parameters values.

        Returns:
            Dict[str, Any]: Algorithm parameters.

        See Also:
            * :func:`niapy.algorithms.algorithm.Algorithm.get_parameters`

        """
        d = super().get_parameters()
        d.update({
            'g0': self.g0,
            'epsilon': self.epsilon
        })
        return d

    def gravity(self, t):
        r"""Get new gravitational constant.

        Args:
            t (int): Time (Current iteration).

        Returns:
            float: New gravitational constant.

        """
        return self.g0 / t

    def init_population(self, task):
        r"""Initialize staring population.

        Args:
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray[float], Dict[str, Any]]:
                1. Initialized population.
                2. Initialized populations fitness/function values.
                3. Additional arguments:
                    * velocities (numpy.ndarray[float]): Velocities

        See Also:
            * :func:`niapy.algorithms.algorithm.Algorithm.init_population`

        """
        population, fitness, _ = super().init_population(task)
        velocities = np.zeros((self.population_size, task.dimension))
        return population, fitness, {'velocities': velocities}

    def run_iteration(self, task, population, population_fitness, best_x, best_fitness, **params):
        r"""Core function of GravitationalSearchAlgorithm algorithm.

        Args:
            task (Task): Optimization task.
            population (numpy.ndarray): Current population.
            population_fitness (numpy.ndarray): Current populations fitness/function values.
            best_x (numpy.ndarray): Global best solution.
            best_fitness (float): Global best fitness/function value.
            **params (Dict[str, Any]): Additional arguments.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, float, Dict[str, Any]]:
                1. New population.
                2. New populations fitness/function values.
                3. New global best solution
                4. New global best solutions fitness/objective value
                5. Additional arguments:
                    * velocities (numpy.ndarray): Velocities.

        """
        velocities = params.pop('velocities')

        ib, iw = np.argmin(population_fitness), np.argmax(population_fitness)
        m = (population_fitness - population_fitness[iw]) / (population_fitness[ib] - population_fitness[iw])
        m = m / np.sum(m)
        forces = np.asarray([[self.gravity((task.iters + 1)) * ((m[i] * m[j]) / (euclidean(population[i], population[j]) + self.epsilon)) * (
                population[j] - population[i]) for j in range(len(m))] for i in range(len(m))])
        total_force = np.sum(self.random((self.population_size, task.dimension)) * forces, axis=1)
        a = total_force.T / (m + self.epsilon)
        velocities = self.random((self.population_size, task.dimension)) * velocities + a.T
        population = np.apply_along_axis(task.repair, 1, population + velocities, self.rng)
        population_fitness = np.apply_along_axis(task.eval, 1, population)
        best_x, best_fitness = self.get_best(population, population_fitness, best_x, best_fitness)
        return population, population_fitness, best_x, best_fitness, {'velocities': velocities}



# encoding=utf8

import numpy as np

from niapy.algorithms.algorithm import Algorithm

__all__ = ['GreyWolfOptimizer']


class GreyWolfOptimizer(Algorithm):
    r"""Implementation of Grey wolf optimizer.

    Algorithm:
        Grey wolf optimizer

    Date:
        2018

    Author:
        Iztok Fister Jr. and Klemen Berkovič

    License:
        MIT

    Reference paper:
        * Mirjalili, Seyedali, Seyed Mohammad Mirjalili, and Andrew Lewis. "Grey wolf optimizer." Advances in engineering software 69 (2014): 46-61.
        * Grey Wolf Optimizer (GWO) source code version 1.0 (MATLAB) from MathWorks

    Attributes:
        Name (List[str]): List of strings representing algorithm names.

    See Also:
        * :class:`niapy.algorithms.Algorithm`

    """

    Name = ['GreyWolfOptimizer', 'GWO']

    @staticmethod
    def info():
        r"""Get algorithm information.

        Returns:
            str: Algorithm information.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""Mirjalili, Seyedali, Seyed Mohammad Mirjalili, and Andrew Lewis. "Grey wolf optimizer." Advances in engineering software 69 (2014): 46-61."""

    def init_population(self, task):
        r"""Initialize population.

        Args:
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, Dict[str, Any]]:
                1. Initialized population.
                2. Initialized populations fitness/function values.
                3. Additional arguments:
                    * alpha (numpy.ndarray): Alpha of the pack (Best solution)
                    * alpha_fitness (float): Best fitness.
                    * beta (numpy.ndarray): Beta of the pack (Second best solution)
                    * beta_fitness (float): Second best fitness.
                    * delta (numpy.ndarray): Delta of the pack (Third best solution)
                    * delta_fitness (float): Third best fitness.

        See Also:
            * :func:`niapy.algorithms.Algorithm.init_population`

        """
        pop, fpop, d = super().init_population(task)
        si = np.argsort(fpop)
        alpha = np.copy(pop[si[0]])
        alpha_fitness = fpop[si[0]]
        beta = np.copy(pop[si[1]])
        beta_fitness = fpop[si[1]]
        delta = np.copy(pop[si[2]])
        delta_fitness = fpop[si[2]]
        d.update({
            'alpha': alpha,
            'alpha_fitness': alpha_fitness,
            'beta': beta,
            'beta_fitness': beta_fitness,
            'delta': delta,
            'delta_fitness': delta_fitness
        })
        return pop, fpop, d

    def run_iteration(self, task, population, population_fitness, best_x, best_fitness, **params):
        r"""Core function of GreyWolfOptimizer algorithm.

        Args:
            task (Task): Optimization task.
            population (numpy.ndarray): Current population.
            population_fitness (numpy.ndarray): Current populations function/fitness values.
            best_x (numpy.ndarray):
            best_fitness (float):
            **params (Dict[str, Any]): Additional arguments.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, float, Dict[str, Any]]:
                1. New population
                2. New population fitness/function values
                3. Additional arguments:
                    * alpha (numpy.ndarray): Alpha of the pack (Best solution)
                    * alpha_fitness (float): Best fitness.
                    * beta (numpy.ndarray): Beta of the pack (Second best solution)
                    * beta_fitness (float): Second best fitness.
                    * delta (numpy.ndarray): Delta of the pack (Third best solution)
                    * delta_fitness (float): Third best fitness.

        """
        alpha = params.pop('alpha')
        alpha_fitness = params.pop('alpha_fitness')
        beta = params.pop('beta')
        beta_fitness = params.pop('beta_fitness')
        delta = params.pop('delta')
        delta_fitness = params.pop('delta_fitness')

        a = 2 - task.evals * (2 / task.max_evals)
        for i, w in enumerate(population):
            a1, c1 = 2 * a * self.random(task.dimension) - a, 2 * self.random(task.dimension)
            x1 = alpha - a1 * np.fabs(c1 * alpha - w)
            a2, c2 = 2 * a * self.random(task.dimension) - a, 2 * self.random(task.dimension)
            x2 = beta - a2 * np.fabs(c2 * beta - w)
            a3, c3 = 2 * a * self.random(task.dimension) - a, 2 * self.random(task.dimension)
            x3 = delta - a3 * np.fabs(c3 * delta - w)
            population[i] = task.repair((x1 + x2 + x3) / 3, rng=self.rng)
            population_fitness[i] = task.eval(population[i])
        for i, f in enumerate(population_fitness):
            if f < alpha_fitness:
                alpha, alpha_fitness = population[i].copy(), f
            elif alpha_fitness < f < beta_fitness:
                beta, beta_fitness = population[i].copy(), f
            elif beta_fitness < f < delta_fitness:
                delta, delta_fitness = population[i].copy(), f
        best_x, best_fitness = self.get_best(alpha, alpha_fitness, best_x, best_fitness)
        return population, population_fitness, best_x, best_fitness, {'alpha': alpha, 'alpha_fitness': alpha_fitness,
                                                                      'beta': beta, 'beta_fitness': beta_fitness,
                                                                      'delta': delta, 'delta_fitness': delta_fitness}
